#!/usr/bin/env python3

import sys

"""
Handle all the error messages in emmer.

In emmer, errors always need to be address to ensure the completely excution. Warnings, on
the other hand, are usually non-critical and can be automatically fixed by emmer.
"""

class Error(Exception):
    """
    Error class

    Arguments:
        code -- Type: str
                error code number in str

    Attributes:
        code -- Type: str
    """
    def __init__(self, code):
        self.code = code


def aftermath(fn):
    """
    What should emmer behave when error occurs?
    1. Print the docstring for ErrorCode __init__ function.
       It is much easier to write detail explanation for each error or
       warning in docstring than in print()
    2. Whether exist the program
       Suppress sys.exit() when running unittest
    """
    def wapper(*args, **kwargs):

        suppress = fn(*args, **kwargs)
        print(fn.__doc__)

        if suppress == False:
            sys.exit()
        return

    return wapper


class ErrorCode1(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 1]]
        Parameter setting error:
        Expect a input directory that contains at least one csv file or a specific
        file name end with ".csv" file.

        Example:
            'emmer/data/data_dir_3'
            'emmer/data/data_dir_3/group_A.csv'
        """
        return(suppress)


class ErrorCode2(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 2]]
        Parameter setting error:
        Expect: 0 < args.z <= 1.
        When set -z as 0.25, a column with more than 1/4 of its elements that are zeros
        will be removed before nominating information-rich features.
        """
        return(suppress)


class ErrorCode3(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 3]]
        Parameter setting error:
        Please set args.z (-z) when using hard filter. Expect: 0 < args.z <= 1.
        When set -z as 0.25, a column with more than 1/4 of its elements that are zeros
        will be removed before nominating information-rich features.
        """
        return(suppress)


class ErrorCode4(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 4]]
        Parameter setting error:
        No feature was selected.
        The information-rich feature selecting thresholds (-u, -l, -t) are too stringent.
        Please consider to relax the thresholds before rerun emmer.harvest

        After successful run emmer.harvest, you can use the "RevisitThreshold" mode in
        emmer.bake to refine theshold selection
        """
        return(suppress)


class ErrorCode5(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 5]]
        Parameter setting error:
        Must set at least one of thresholds (-u: upper; -l: lower) for information-rich
        feature selection.
        """
        return(suppress)


class ErrorCode6(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 6]]
        Parameter setting error:
        Expect detection limit args.d (-d) to be less than the greatest number in your
        input matrix. Please check your detection limit setting and re-run the program.
        """
        return(suppress)


class ErrorCode7(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 7]]
        Developer error:
        Expect to pass 'expect' argument when using EvaluateInput to evaulate the
        input code.

        Please note that not all of EvaluateInput evaluation needs a 'expect' argument.
        """
        return(suppress)


class ErrorCode8(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 8]]
        Parameter setting error:
        Expect a emmer.harvest-gereated csv file.

        The file usually stored under "output" folder with the keyword "projection"
        in its file name.
        """
        return(suppress)


class ErrorCode9(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 9]]
        Parameter setting error:
        Expect a emmer.harvest-gereated csv file that contain percent explain information
        for each PC.
        """
        return(suppress)


class ErrorCode10(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 10]]
        Unexpected user response error
        The number of the elements in the input, which should be separated by semicolon when
        user enter the input, does not match the number of attributes that needs input.

        For example, this error will arise when emmer.bake ask the user to define color for
        two groups of data on the PCA plot, but the number of color inputed by the user are
        not equal to two.
        """
        return(suppress)


class ErrorCode11(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 11]]
        Unexpected input error:
        Expect input PCA coordinate file to contain the coordinates for at least
        two dimensions.

        If you are using a emmer.harvest output file, please consider to relax
        your information-rich feature choosing thresholds. You can also run
        'RevisitThreshold' mode in emmer.bake to select a better threshold settings.
        """
        return(suppress)


class ErrorCode12(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 12]]
        Parameter setting error:
        Expect a csv file for args.p (-p)
        You can find the file for -p in output/*information_rich_features_summary.csv
        after running 'Bifurication' mode in emmer.bake
        """
        return(suppress)


class ErrorCode13(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 13]]
        Parameter setting error:
        Expect a directory that store all the *detail_vNE.csv generated from
        emmer.harvest

        Those files can are stored under "~/output/detial_vNE"
        """
        return(suppress)


class ErrorCode14(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 14]]
        Parameter setting error:
        Expect a directory that store all the *pre_filterd_data.csv generated from
        emmer.harvest

        Those files can are stored under "~/output/pre_filterd_data"
        """
        return(suppress)


class ErrorCode15(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 15]]
        Parameter setting error:
        direction designated by args.i (-i) should contain all and only the files
        that were used to generate the files in args.v (-v).
        """
        return(suppress)


class ErrorCode16(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 16]]
        Parameter setting error:
        Expect a single csv file that contains information for scaling the data matrix for calculating the
        PCA coordinate. This file is generated by emmer.harvest when set -n as True.

        Hint: the file stored under output/ that contains '__data_colstd' in file name.
        """
        return(suppress)


class ErrorCode17(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 17]]
        Parameter setting error:
        Expected to have three elements that are separated by comma from input argument.

        Example:
            -t '2,1,1'
            -u '3,2,1'
            -l '2,2,0'
        """
        return(suppress)


class ErrorCode18(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 18]]
        Parameter setting error:
        Expected to have at least one of these three setting (-t, -u, -l) when set -m
        as "RevisitThreshold"
        """
        return(suppress)


class ErrorCode19(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 19]]
        Parameter setting error:
        The first two elements in input should be a positive number that is greater than 0.

        These two numbers represent x-fold of standard deviation from the mean of the von
        Neumann entropy calculated from systemically remove one column (feature) at a time.
        Expect x > 0.
        """
        return(suppress)


class ErrorCode20(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 20]]
        Parameter setting error:
        The first two elements in input should be a positive number that is greater
        than 0.

        The numbers represent x-fold of standard deviation from the mean of the
        von Neumann entropy calculated from systemically remove one column (feature)
        at a time. Expect x > 0.
        """
        return(suppress)


class ErrorCode21(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 21]]
        Parameter setting error:
        Input directory (-i) should contains at least two .csv file. Please make
        sure you enter the correct directory path.

        Please refer to '/data/data_dir_3' as an example for the input dictory.
        """
        return(suppress)


class ErrorCode22(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 22]]
        Parameter setting error:
        The thrid element in arguments -u, -l, and -t represents the increment.
        Elements are separated by comma. Example: -u <max>,<min>,<increment>
        Expect the residual = 0 when divide the (max - min)/increment

        Please note the increment of -t setting can only be an interger.

        Example:
            -u 2,1,1
            -u 2,1,0.5
        """
        return(suppress)


class ErrorCode23(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 23]]
        Parameter setting error:
        Expect input as a an interger or a float.
        """
        return(suppress)


class ErrorCode24(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 24]]
        Parameter setting error:
        The thrid element in arguments -u, -l, and -t represents the increment.
        Base on your setting of the first and the second elements in this argument,
        EMMER except a number that is greater than zero.

        Example:
            -u '3,1,1'

        For the example above, emmer.bake will test three different -u setting:
        1, 2, and 3
        """
        return(suppress)


class ErrorCode25(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 25]]
        Parameter setting error:
        The thrid element in arguments -u, -l, and -t represents the increment.
        Base on your setting of the first and the second elements in this argument,
        EMMER except a number that is less than the first elements.

        Example:
            -u '3,1,1'
            -l '3,1,0.5'
        """
        return(suppress)


class ErrorCode26(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 26]]
        Parameter setting error:
        The thrid element in arguments -u, -l, and -t represents the increment.
        Base on your setting of the first and the second elements in this argument,
        EMMER except a number that equals to zero.

        Example:
            -u 2,2,0
        """
        return(suppress)


class ErrorCode27(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 27]]
        Parameter setting error:
        No information-rich feature was selected in any combination of the input
        threshold settings (-u, -l, -t)

        Suggestion:
            Select least stringent thresholds that include the original setting
            when user first run emmer.harvest
        """
        return(suppress)


class ErrorCode28(Error):

    @aftermath
    def __init__(self, suppress):
        """
        Unexpected user response:
        Expect matplotlib.colors (https://matplotlib.org/api/colors_api.html; for example: red,
        blue) or HEX color (#000000)

        Example:
        Assuming you have two groups, group_A and group_B. You want to assign red and blue to
        group_A and group_B respectively. You can enter:
        red;blue
        #ff0000;#0000ff
        #FF0000;#0000FF
        """
        return(suppress)


class ErrorCode29(Error):

    @aftermath
    def __init__(self, suppress = False):
        """
        Input file format error:
        Expect a csv file that has ['PC1', 'PC2'] or ['PC1', 'PC2', 'PC3'] as column names
        in the first two or three columns.

        The order of column names matters. A file that has the column name in ['PC2', 'PC1']
        order will also trigger this error code
        """
        return(suppress)


class ErrorCode30(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 30]]
        Parameter setting error:
        Expect a csv file to have ['group', 'individual', 'edge_color', 'fill_color', 'shape'] as
        column names.

        If you are not sure how to prepare -p file, you don't have use -p argument when running
        emmer.bake. The program will guild you through the process and prepare a -p file for your
        future analysis.

        Example:
            python3 -m emmer.bake -m 'Individual' -i <filtered_infoRich__PCA_coordinates.csv>
            python3 -m emmer.bake -m 'Permanova' -i <filtered_infoRich__PCA_coordinates.csv>
            python3 -m emmer.bake -m 'Reproducibility' -b 20 -i <information_rich_features_summary.csv>
            python3 -m emmer.bake -m 'RevisitThreshold' -u <2.5,1.5,0.25> -l <2.5,1.5,0.25> -v <output/detail_vNE/> -i <output/filtered_data/>
            python3 -m emmer.bake -m 'Bifurication' -i <filtered_data/> -p <information_rich_features_summary.csv>


        Group:
            Expect a csv file that has ['group'] as column names. In emmer.bake, the row of the
            input matrix represent each individual in each original input file for emmer. 'Group'
            are the same as orginal input file names.

        Individaul
            Expect a csv file that has ['individual'] as column names. In Retrospect, the row of the
            input matrix represent each individual in each original input file for emmer. 'individual'
            are the same as rows in orginal input files.

        edge_color:
            Expect a csv file that has ['edge_color'] as column names.

        fill_color:
            Expect a csv file that has ['fill_color'] as column names.

        Shape:
            Expect a csv file that has ['shape'] as column names.
        """
        return(suppress)


class ErrorCode31(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 31]]
        Developer error:
        Do not expect to pass 'expect' argument when using EvaluateInput to evaluateColor
        the input code.

        Please use the default expect setting. Let EvaluateInput.expect = 0
        """
        return(suppress)


class ErrorCode32(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 32]]
        Parameter setting error:
        Expect a single csv file that contains PCA coordinate generated by emmer.harvest

        Hint: the file stored under output/ that contains 'PCA_coordinates' in file name.
        """
        return(suppress)


class ErrorCode33(Error): #

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 33]]
        Parameter setting error:
        Expect matplotlib marker name (https://matplotlib.org/api/markers_api.html)
        """
        return(suppress)


class ErrorCode34(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 34]]
        Parameter setting error:
        Expect a single csv file that contains transformation matrix for generating the PCA coordinate.
        This file is generated by emmer.harvest

        Hint: the file stored under output/ that contains 'transformation_matrix' in file name.
        """
        return(suppress)


class ErrorCode35(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 35]]
        Parameter setting error:
        Expect a single csv file that contains information for scaling the data matrix for calculating the
        PCA coordinate. This file is generated by emmer.harvest

        Hint: the file stored under output/ that contains '__data_colmean' in file name.
        """
        return(suppress)


class ErrorCode36(Error):

    @aftermath
    def __init__(self, suppress):
        """
        Unexpected user response:
        Please choose between 'Group' or 'Individaul'
        ['Group', 'Individual']

        or use one the corresponding option number
        ['1', '2']
        """
        return(suppress)


class ErrorCode37(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 37]]
        Unexpected user response:
        Expect to have more than one cluster when running 'Permanova' mode
        """
        return(suppress)


class ErrorCode38(Error):

    @aftermath
    def __init__(self, suppress):
        """
        Input file naming error:
        Please avoid having "__" in the input csv file name.
        Having "__" will interfere with emmer.harvest and emmer.bake data handling

        Please refer to /data/data_dir_3 as an example for the input file format.
        """
        return(suppress)


class ErrorCode39(Error):

    @aftermath
    def __init__(self, suppress = False):
        """
        [[Error code 39]]
        Input file format error:
        Please avoid having "__" in row names of the input csv file.
        Having "__" will interfere with EMMER data handling

        Please refer to /data/data_dir_3/group_A.csv as example for
        emmer.harvest input file
        """
        return(suppress)


class ErrorCode40(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 40]]
        Input file format error:
        Expect column headers of the input mapping file to be "file_name"
        for the first column and "sample_id" for the second column.

        Please refer to /data/sow_test_dir_2/correct_mapping_file.csv as
        an example for the input mapping file.
        """
        return(suppress)



class ErrorCode41(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 41]]
        Expect input matrix (from args.i; -i) to have more than one row (sample).
        """
        return(suppress)


class ErrorCode42(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 42]]
        Parameter setting error:
        Input 'transformation_matrix' and '__data_colmean' might be from different analyses. Please
        choose the 'transformation_matrix' file and '__data_colmean' file generated from the same
        emmer.harvest run

        To pass:
        Make sure both file have exactly the same set of feature names (as rowname in
        'transformation_matrix'; as colname in '__data_colmean') and they are at the same order
        """
        return(suppress)


class ErrorCode43(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 43]]
        Parameter setting error:
        Expect new observation (args.x; -x)
        Expect a input directory that contains at least one csv file or a specific
        file name end with ".csv" file.

        Example:
            'emmer/data/data_dir_3'
            'emmer/data/data_dir_3/group_A.csv'
        """
        return(suppress)


class ErrorCode44(Error):  # TODO need unittest

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 44]]
        Log file already exist; raise error and stop emmer to avoid accidently remove
        previous log file without knowing.

        To prevent this error when running multiple emmer analyses in parallele, please
        provide a unique user-defined tag for each analysis via -o (args.o).
        """
        return(suppress)


class ErrorCode45(Error):

    @aftermath
    def __init__(self, suppress = False):
        """
        [[Error code 45]]
        Input file should not contains null value.
        """
        return(suppress)


class ErrorCode46(Error):       # TODO: unittest

    @aftermath
    def __init__(self, suppress = False):
        """
        [[Error code 46]]
        Expect a numeric input matrix. At least one of your column contains non-numeric
        values. Please check your input file(s).
        """
        return(suppress)


class ErrorCode47(Error):

    @aftermath
    def __init__(self, suppress = False):
        """
        [[Error code 47]]
        Parameter setting error:
        Number of CPU dedicated in this analysis should not exceed the number of
        CPU in your computer.
        """
        return(suppress)


class ErrorCode48(Error):

    @aftermath
    def __init__(self, suppress):
        """
        [[Error code 48]]
        Parameter setting error:
        Input '__data_colstd' and '__data_colmean' might be from different analyses. Please
        choose the '__data_colstd' file and '__data_colmean' file generated from the same
        emmer.harvest run.

        To pass:
        Make sure both file have exactly the same set of feature names and they are at the
        same order.
        """
        return(suppress)



#!/usr/bin/env python3

## usage:
# at a level above emmer/
# python3 -m emmer.test.test_technical

from ..troubleshoot.err.error import ErrorCode23
from ..toolbox.technical import *
import unittest
import numpy as np
import pandas as pd
import sys


class TestFlattern(unittest.TestCase):

    def test_flattern(self):
        print('\ntest_flattern:')
        print('        list_of_list = [[1, 2], [3, 4]]')
        list_of_list = [[1, 2], [3, 4]]

        my_result = flattern(list_of_list)
        expected_result = [1, 2, 3, 4]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


class TestEmptyNumpyArray(unittest.TestCase):

    def test_emptyNumpyArray(self):
        print('\ntest_emptyNumpyArray:')
        print('        Case 1: confirm the function generate a empty array')
        print('        nrow = 3')
        nrow = 3
        print('        ncol = 4')
        ncol = 4

        np_array = emptyNumpyArray(nrow = 3, ncol = 4)

        ## Case 1: confirm the function generate a empty array
        df = pd.DataFrame(data = np_array)
        my_result = sum(flattern(df.isnull().values.tolist()))
        expected_result = 12
        self.assertEqual(my_result, expected_result)

        ## Case 2: confirm deminsion
        print('        ---------------------------------------------------')
        print('        Case 2: confirm deminsion')
        my_result = df.shape[0]
        expected_result = 3
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestToFloat(unittest.TestCase):

    def test_toFloat(self):
        print('\ntest_toFloat:')
        print('        Case 1: match expected input')
        print('        number_in_str = 9')
        number_in_str = 9

        ## Case 1: match expected input
        my_result = toFloat(number_in_str, suppress = False)
        expected_result = 9
        self.assertEqual(my_result, expected_result)

        ## Case 2: unexpected input; arise error msg
        print('        ---------------------------------------------------')
        print('        Case 2: unexpected input; arise error message')
        print('        number_in_str = "A"')
        number_in_str = 'A'

        with self.assertRaises(ErrorCode23):
            output = toFloat(number_in_str, suppress = True)
        print('===========================================================')


class TestFloatRange(unittest.TestCase):

    def test_floatRange(self):
        ## Case 1: normal input
        print('\ntest_floatRange:')
        print('        Case 1: normal input')
        print('        input_tuple = (3, 1, 0.5)')
        input_tuple = (3, 1, 0.5)

        my_result = floatRange(input_tuple)
        expected_result = [1, 1.5, 2, 2.5, 3]
        self.assertEqual(my_result, expected_result)

        ## Case 2: max == min
        print('        ---------------------------------------------------')
        print('        Case 2: max == min')
        print('        input_tuple = (2, 2, 0)')
        input_tuple = (2, 2, 0)

        my_result = floatRange(input_tuple)
        expected_result = [2]
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestAddElementsInList(unittest.TestCase):

    def test_addElementsInList(self):
        print('\ntest_addElementsInList:')
        print('        list_1 = [1, 2, 3]')
        list_1 = [1, 2, 3]
        print('        list_2 = [3, 2, 1]')
        list_2 = [3, 2, 1]

        my_result = addElementsInList(list_1 = list_1, list_2 = list_2)
        expected_result = [4, 4, 4]
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestDualAssignment(unittest.TestCase):

    def test_dualAssignment(self):
        print('\ntest_addElementsInList:')
        print('        dataframe = pd.DataFrame([[1,2,3,4],[2,3,2,3],[1,3,2,4]], columns = ["A", "A-B", "B", "A-C"])')
        dataframe = pd.DataFrame([[1,2,3,4],[2,3,2,3],[1,3,2,4]], columns = ["A", "A-B", "B", "A-C"])
        print('        sep = "-"')
        sep = "-"

        print(dataframe)
        my_result = dualAssignment(dataframe = dataframe, sep = sep)
        print(my_result)
        expected_result = {'A': [7, 8, 8], 'B': [5, 5, 5], 'C': [4, 3, 4]}
        self.assertEqual(my_result, expected_result)
        print('===========================================================')
        

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_read

from ..main.basic.read import RawDataImport, GetFiles, MergeTargetedFiles, RetrospectDataImport
from ..troubleshoot.err.error import ErrorCode1, ErrorCode9, ErrorCode11, ErrorCode29, ErrorCode30, ErrorCode38, ErrorCode39, ErrorCode40, ErrorCode45

from pandas.util.testing import assert_frame_equal
import unittest
import pandas
import numpy
import os


class TestGetFiles(unittest.TestCase):

    def test_GetFiles(self):
        print('\ntest_GetFiles:')
        print('        case 1: get the name of files from the user-defined directory')
        #input_dir = 'emmer/data/data_dir_2/'
        input_dir = 'emmer/data/data_dir_3/'
        multiple_input = GetFiles(input_dir = input_dir, suppress = False, second_chance = False)
        my_result = sorted(list(multiple_input.input_files))

        expected_result = [os.path.join(input_dir, 'group_A.csv'), os.path.join(input_dir, 'group_B.csv')]
        self.assertListEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        case 2: test error handling (no csv in user-defined directory)')
        input_dir = 'emmer/data/data_dir_5/'
        with self.assertRaises(ErrorCode1):
            multiple_input = GetFiles(input_dir = input_dir, suppress = True, second_chance = False)
        print('===========================================================')


class TestRawDataImport(unittest.TestCase):

    def test_readCSV(self):
        print('\ntest_readCSV:')
        print('        case 1: read csv files')
        input_matrix = RawDataImport(file_name = 'emmer/data/test_case_1.csv', for_merging_file = False, suppress = False)
        input_matrix.readCSV()

        data = {'col1':[10, 21, 42, 21, 45 ,0],
                'col2':[1, 4, 10, 6, 14, 14],
                'col3':[2, 4, 2, 3, 1, 13],
                'col4':[0, 1, 0, 0, 0, 0],
                'col5':[12, 17, 18, 30, 14, 1],
                'col6':[50, 30, 26, 23, 25, 4]}
        expected_result = pandas.DataFrame(data, index =['sample1', 'sample2', 'sample3',
                                                       'sample4', 'sample5', 'sample7'])
        assert_frame_equal(input_matrix.data, expected_result)

        print('        ---------------------------------------------------')
        print('        case 2: test error handling')
        print('             2.1: file name should not contain "__"')
        with self.assertRaises(ErrorCode38):
            input_matrix = RawDataImport(file_name = 'emmer/data/data_dir_6/test__case_2.csv', for_merging_file = False, suppress = True)
            input_matrix.readCSV()

        print('        ---------------------------------------------------')
        print('             2.2: row name should not contain "__"')
        with self.assertRaises(ErrorCode39):
            input_matrix = RawDataImport(file_name = 'emmer/data/data_dir_6/test_case_1.csv', for_merging_file = False, suppress = True)
            input_matrix.readCSV()

        print('        ---------------------------------------------------')
        print('             2.3: input file name should ends with ".csv"')
        with self.assertRaises(ErrorCode1):
            input_matrix = RawDataImport(file_name = 'emmer/data/sow_test_dir_2/targert_file_1.txt', for_merging_file = False, suppress = True)
            input_matrix.readCSV()

        print('        ---------------------------------------------------')
        print('             2.4: file not found')
        with self.assertRaises(ErrorCode1):
            input_matrix = RawDataImport(file_name = 'emmer/data/data_dir_5/test.csv', for_merging_file = False, suppress = True)
            input_matrix.readCSV()
        print('===========================================================')


    def test_relativeAbundance(self):
        print('\ntest_RawDataImport.relativeAbundance:')
        input_matrix = RawDataImport(file_name = 'emmer/data/test_case_1.csv')
        input_matrix.readCSV()
        input_matrix.relativeAbundance()
        my_result = list(pandas.DataFrame.sum(input_matrix.data, axis = 1))
            # set axis = 1 to sum the rows

        expected_result = [1, 1, 1, 1, 1, 1]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


    def test_detectionLimit(self):
        print('\ntest_RawDataImport.detectionLimit:')
        print('        case 1: complete entry that below the detection limit.\nThen remove empty row and column.')
        input_matrix = RawDataImport(file_name = 'emmer/data/test_case_1.csv', for_merging_file = False, suppress = True)
        input_matrix.readCSV()
        input_matrix.relativeAbundance()
        input_matrix.detectionLimit(detection_limit = 0.4)

        data = {'col1':[0, 0.428571, 0.454545 ,0],
                'col2':[0, 0, 0, 0.4375],
                'col3':[0, 0, 0, 0.40625],
                'col6':[0.666667, 0, 0, 0]}
        expected_result = pandas.DataFrame(data, index =['sample1', 'sample3', 'sample5', 'sample7'])
        assert_frame_equal(input_matrix.data, expected_result)
        print('===========================================================')


    def test_hardFilter(self):
        print('\ntest_RawDataImport.detectionLimit:')
        print('        case 1: complete entry that below the detection limit. Then remove empty row and column.')
        input_matrix = RawDataImport(file_name = 'emmer/data/test_case_1.csv', for_merging_file = False, suppress = True)
        input_matrix.readCSV()
        input_matrix.relativeAbundance()
        input_matrix.detectionLimit(detection_limit = 0.2)
        input_matrix.hardFilter(zero_tolerance_level = 0.5)
        my_result = list(input_matrix.feature_names)

        expected_result = ['col1', 'col6']
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


class TestMergeTargetedFiles(unittest.TestCase):

    def test_getMap(self):
        print('\ntest_MergeTargetedFiles.getMap:')
        print('        case 1: get mapping file')
        print('             1.1: correct mapping file')
        mapping_file = "emmer/data/sow_test_dir_2/correct_mapping_file.csv"
        target_file_dir = "emmer/data/sow_test_dir_2"
        suppress = False

        merge_target_files = MergeTargetedFiles(mapping_file = mapping_file, target_file_dir = target_file_dir,
                                                separate = '\t', suppress = suppress)
        merge_target_files.getMap()

        print(merge_target_files.map.columns.values)
        my_result = list(merge_target_files.map.columns.values)

        expected_result = ['file_name', 'sample_id', 'additional_note']
        self.assertListEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             1.2: raise error when handling mapping file that has incorrect format')
        mapping_file = "emmer/data/sow_test_dir_2/incorrect_mapping_file.csv"
        target_file_dir = "emmer/data/sow_test_dir_2"
        suppress = True

        with self.assertRaises(ErrorCode40):
            merge_target_files = MergeTargetedFiles(mapping_file = mapping_file, target_file_dir = target_file_dir,
                                                    separate = '\t', suppress = suppress)
            merge_target_files.getMap()
        print('===========================================================')


        print('\ntest_MergeTargetedFiles.getTargetFile:')
        mapping_file = "emmer/data/sow_test_dir_2/correct_mapping_file.csv"
        target_file_dir = "emmer/data/sow_test_dir_2"
        target_file = "emmer/data/sow_test_dir_2/targert_file_1.txt"
        suppress = False

        merge_target_files = MergeTargetedFiles(mapping_file = mapping_file, target_file_dir = target_file_dir,
                                                separate = '\t', suppress = suppress)
        target = merge_target_files.getTargetFile(targeted_file = target_file, col_name = 'A')
        my_result = list(target.shape)

        expected_result = [3,2]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


        print('\ntest_MergeTargetedFiles.mergeFile:')
        mapping_file = "emmer/data/sow_test_dir_2/correct_mapping_file.csv"
        target_file_dir = "emmer/data/sow_test_dir_2"
        suppress = False

        merged = MergeTargetedFiles(mapping_file = mapping_file, target_file_dir = target_file_dir,
                                                separate = '\t', suppress = suppress)
        merged.mergeFile()
        print(merged.merged_target_file)
        my_result = list(merged.merged_target_file.index)

        expected_result = ['feature_1', 'feature_2', 'feature_3', 'feature_4']
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


class TestRetrospectDataImport(unittest.TestCase):

    def test_RetrospectDataImport(self):
        print('\ntest_RetrospectDataImport:')
        print('        case 1: __init__ error handling')
        print('             1.1: input file contains null value')
        with self.assertRaises(ErrorCode45):
            imported_bake_data = RetrospectDataImport(file_name = 'emmer/data/problem_maker/read_module/information_rich_features_summary_w_NA.csv', type = 'reproducibility', dimension = 'n', suppress = True)


        print('        ---------------------------------------------------')
        print('             1.2: file not found')
        with self.assertRaises(ErrorCode1):
            imported_bake_data = RetrospectDataImport(file_name = 'emmer/data/problem_maker/read_module/a_file_that_never_exist.csv', type = 'coordinate', dimension = 'n', suppress = True)


        print('        ---------------------------------------------------')
        print('        case 2. "coordinate" error handling')
        print('             2.2: input only contains coordinate for one PC')
        with self.assertRaises(ErrorCode11):
            imported_bake_data = RetrospectDataImport(file_name = 'emmer/data/problem_maker/read_module/PC1_only_coordinates.csv', type = 'coordinate', dimension = 'n', suppress = True)


        print('        ---------------------------------------------------')
        print('        case 3. "reproducibility"')
        print('             3.1: expect to work')
        print('             3.1.1.: summarized reproducibilty output')
        print('                     row: information-rich taxa; column: group')
        imported_repro_data = RetrospectDataImport(file_name = 'emmer/data/bake_data_dir_8/case1.csv', type = 'reproducibility', suppress = False)
        my_result = numpy.mean(numpy.array(imported_repro_data.reproducibility))
        expected_result = 0.40625
        self.assertAlmostEqual(my_result, expected_result)


        print('             3.1.2.: reproducibilty output from a single input file')
        print('                     row: information-rich taxa; column: ["feature_name", "occurrence", "repreducibility (%)"]')
        imported_repro_data = RetrospectDataImport(file_name = 'emmer/data/bake_data_dir_8/case2.csv', type = 'reproducibility', suppress = False)
        my_result = numpy.mean(numpy.array(imported_repro_data.reproducibility))
        expected_result = 75.6302521008403
        self.assertAlmostEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('        case 4: unexpected input format when type == "data_color_shape_and_fill"')
        print('             4.1: incorrect column header')
        with self.assertRaises(ErrorCode29):
            imported_bake_data = RetrospectDataImport(file_name = 'emmer/data/problem_maker/read_module/incorrect_header_data_color_shape_and_fill.csv', type = 'data_color_shape_and_fill', dimension = 'n', suppress = True)

        print('        ---------------------------------------------------')
        print('             4.2: no column header named edge_color')
        with self.assertRaises(ErrorCode30):
            imported_bake_data = RetrospectDataImport(file_name = 'emmer/data/problem_maker/read_module/no_edge_color.csv', type = 'data_color_shape_and_fill', dimension = 'n', suppress = True)


        print('        ---------------------------------------------------')
        print('        case 5: unexpected input file for "precent_explained"')
        with self.assertRaises(ErrorCode9):
            imported_bake_data = RetrospectDataImport(file_name = 'emmer/data/problem_maker/read_module/incorrect_precent_explained.csv', type = 'precent_explained', dimension = 'n', suppress = True)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from ...main.basic.read import RawDataImport, RetrospectDataImport, GetFiles
from ...toolbox.technical import emptyNumpyArray
from ...troubleshoot.err.error import *

import pandas
import numpy
import sys
import os


"""
Support emmer.bake Projection mode

Plot new observation onto the existing PCA space. Generate a PCA with original data
points. Does not allow for customizing figure aesthetics (leave this part to emmer.bake
Individaul mode).
"""

class ProjectionArgs:
    """
    Take common arguments for Projection modes when running bake modules

    Argument:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        current_wd -- Type: str
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        #silence -- Type: boolean

    Attributes:
        original_pca_coordinate -- Type: RetrospectDataImport class object (pandas Dataframe)
                                   PCA coordinates for the original datasets.
                                   Without addition plot aesthetics setting (color, shape, etc.)
        input -- Type: list
                 A list of input file names for new observations
        v_df -- Type: Projection.V_df (pandas DataFrame) # to update
        colmean_df -- Type: RawDataImport object Projection.colmean_df (pandas DataFrame)
        normalize -- Type: boolean. Set as True when -s2 exist
        new_obs_merged -- Type: pandas DataFrame
                          new observations
    """

    def __init__(self, args, current_wd, suppress):


        ##--args.i--##
        try:
            if not args.i:
                raise Error(code = '32')
        except Error as e:
            raise ErrorCode32(suppress = suppress) from e

        try:
            if args.i.endswith('.csv') == True:
                self.original_coordinate = RawDataImport(file_name = os.path.join(current_wd, args.i), for_merging_file = True, suppress = suppress)
                self.original_coordinate.readCSV()
            else:
                raise Error(code = '32')
        except Error as e:       # could use FileNotFoundError
            raise ErrorCode32(suppress = suppress) from e


        ##--args.v--##
        try:
            if not args.v:
                raise Error(code = '34')
        except Error as e:
            raise ErrorCode34(suppress = suppress) from e


        try:
            if args.v.endswith('.csv') == True:
                self.v_df = RetrospectDataImport(file_name = os.path.join(current_wd, args.v), type = 'v_df', dimension = 'n', suppress = suppress).v_df
            else:
                raise Error(code = '34')
        except Error as e:        # could use FileNotFoundError
            raise ErrorCode34(suppress = suppress) from e


        ##--args.s1--##
        try:
            if not args.s1:
            #if not args.s:
                raise Error(code = '35')
        except Error as e:
            raise ErrorCode35(suppress = suppress) from e

        try:
            #if args.s.endswith('.csv') == True:
            #    self.scaler_df = RawDataImport(file_name = os.path.join(current_wd, args.s), for_merging_file = True, suppress = suppress)
            #    self.scaler_df.readCSV()
            if args.s1.endswith('.csv') == True:
                self.colmean_df = RawDataImport(file_name = os.path.join(current_wd, args.s1), for_merging_file = True, suppress = suppress)
                self.colmean_df.readCSV()
            else:
                raise Error(code = '35')
        except Error as e:       # could use FileNotFoundError
            raise ErrorCode35(suppress = suppress) from e


        # make sure args.v and arg.s1 have the same features names and at the same order
        try:
            if list(self.colmean_df.feature_names) != list(self.v_df.index.values):
                # self.projection_args.v_df.index.values is feature_names
                raise Error(code = '42')
        except Error as e:
            raise ErrorCode42(suppress = suppress) from e


        ##--args.s2--##
        if args.s2:
            try:
                if args.s2.endswith('.csv') == True:
                    self.colstd_df = RawDataImport(file_name = os.path.join(current_wd, args.s2), for_merging_file = True, suppress = suppress)
                    self.colstd_df.readCSV()
                else:
                    raise Error(code = '16')
            except Error as e:       # could use FileNotFoundError
                raise ErrorCode16(suppress = suppress) from e

            # make sure args.s1 and arg.s2 have the same features names and at the same order
            try:
                if list(self.colmean_df.feature_names) != list(self.colstd_df.feature_names):
                    raise Error(code = '48')
            except Error as e:
                raise ErrorCode48(suppress = suppress) from e

            self.normalize = True
        else:
            self.normalize = False


        ##--args.x--##
        try:
            if not args.x:
                raise Error(code = '43')
        except Error as e:
            raise ErrorCode43(suppress = suppress) from e

        try:
            if os.path.isdir(os.path.join(current_wd,args.x)) == True:
                self.input = GetFiles(os.path.join(current_wd, args.x)).input_files
            elif args.x.endswith('.csv') == True:
                self.input = [os.path.join(current_wd, args.x)]
            else:
                raise Error(code = '43')
        except Error as e:
            raise ErrorCode43(suppress = suppress) from e


        ## read file and merge into a single dataframe
        for f in range(len(self.input)):
            print(self.input[f])
            new_obs = RawDataImport(self.input[f], for_merging_file = False, suppress = suppress)
            new_obs.readCSV()

            index_tag = new_obs.basename.replace(".csv", "__")
            new_obs.data.index = [index_tag + element for element in new_obs.sample_id]
            new_obs.sample_id = new_obs.data.index

            if args.r:
                new_obs.relativeAbundance()

            if len(self.input) == 0 or f == 0:
                self.new_obs_merged = new_obs.data
            else:
                self.new_obs_merged = pandas.concat([self.new_obs_merged, new_obs.data], axis = 0, sort = True)


        ## make sure the order of column are the same as colmean_df
        new_obs_merged_ordered = pandas.DataFrame(data = emptyNumpyArray(nrow = self.new_obs_merged.shape[0], ncol = len(self.colmean_df.feature_names)),
                                                  columns = self.colmean_df.feature_names, index = self.new_obs_merged.index.values)

        for c in range(len(list(self.colmean_df.feature_names))):
            in_merged_data = self.colmean_df.feature_names[c] in self.new_obs_merged.columns.values
            if in_merged_data == True:
                new_obs_merged_ordered[[self.colmean_df.feature_names[c]]] = self.new_obs_merged[[self.colmean_df.feature_names[c]]]
            else:
                new_obs_merged_ordered[[self.colmean_df.feature_names[c]]] = 0
                ## TODO: need unittest; raise WarningCode?

        self.new_obs_merged_ordered = new_obs_merged_ordered.fillna(0)


def projectNew(args, current_wd, retrospect_dir, output_file_tag, suppress):
    """
    write something...
    """
    projection_args = ProjectionArgs(args = args, current_wd = current_wd, suppress = suppress)

    ## minus colmean_df and matmul with V
    nrow = projection_args.new_obs_merged_ordered.shape[0]
    new_observation_df = projection_args.new_obs_merged_ordered
    colmean = numpy.array(projection_args.colmean_df.data)
    V = numpy.array(projection_args.v_df)

    ncol_V = numpy.size(V, 1)
    PC_list = list(['PC' + str(i) for i in range(1 , (ncol_V + 1))])
    mean_centered_matrix = numpy.array(new_observation_df) - numpy.tile(colmean, (nrow, 1))

    if projection_args.normalize == True:
        colstd = numpy.array(projection_args.colstd_df.data.iloc[0, ])
        scaled_mean_centered_matrix = numpy.matmul(mean_centered_matrix, numpy.diag(1/colstd))   ### FIXME  ##PC order?
        projection = numpy.matmul(scaled_mean_centered_matrix, V)
    else:
        projection = numpy.matmul(mean_centered_matrix, V)

    projection_df = pandas.DataFrame(projection, columns = PC_list, index = new_observation_df.index.values)
    #print(projection_df.columns.values)


    ## TODO: PC1, PC10, PC11, PC2 ... need reorder
    output_file_name = os.path.join(retrospect_dir, (output_file_tag + '__retrospect_project_new.csv'))
    projection_df.to_csv(output_file_name)

    ## merge the coordinates from new and original dataset
    original_and_new_coordinates_df = pandas.concat([projection_args.original_coordinate.data, projection_df], axis = 0, sort = True)
    original_and_new_coordinates_df = original_and_new_coordinates_df.fillna(0)

    output_file_name = os.path.join(retrospect_dir, (output_file_tag + '__retrospect_both_new_and_original_coordinates.csv'))
    original_and_new_coordinates_df.to_csv(output_file_name)

    return(original_and_new_coordinates_df)



#!/usr/bin/env python3

from ..basic.math import NonDesityMatrix
from ..basic.read import RawDataImport
from ...troubleshoot.warn.warning import *
from ...troubleshoot.err.error import *
from ...toolbox.recorder import UpdateNoteBook

from multiprocessing import Pool
from tqdm import tqdm

import pandas
import numpy
import time
import sys
import os

"""
This module is the basic functional unit in EMMER package. The objective of this module
is to import single dataset (a .csv file) and to select information-rich feature.


MinusOneVNE:
Systematically remove one feature from data matrix at each iteration. This class contains
functions that supports for MinDataLostFilter and InfoRichCalling

InfoRichCalling:
This class contains functions for selecting information-rich featue based on MinusOneVNE
output and selecting thresholds. Future expansion might inculde additional options for
selection criteria.

reproducibility() and reproducibility_summary():
These two functions are used to calculate and summarize the information-rich feature
call reproducibility by jackknifting data matrix (systematically remove one sample from
the data matrix at each iteration.)

Kernal:
Smallest functional unit in EMMER_vNE package. The functions in this class takes in a data
matrix (a .csv file) and filter based on user-defined parameters. Then select information-rich
features according to the user-defined threshold. Report reproducibilty upon user's request.
"""

class MinusOneVNE:
    """
    For a given RawDataImport.data matrix. Remove one feature (column) at a time and
    calculate the von Neumann entropy for the remaining matrix.

    To speed up the computation process MinusOneVNE.data is defined as a numpy.array
    """

    def __init__(self, data, normalize, feature_names, num_cpu):
        self.data = numpy.array(data)
        self.feature_names = feature_names
        self.num_cpu = num_cpu
        self.normalize = normalize

    def minusOneVNE(self, col_to_remove):
        """
        Remove specific feature from the data matrix and calculate the von Neumann entropy
        from the remaining matrix.
        """
        subset = numpy.array(self.data)
        # delete column by index
        subset = numpy.delete(subset, col_to_remove, axis = 1)

        minusOneVNE = [col_to_remove, self.feature_names[col_to_remove], NonDesityMatrix(subset, normalize = self.normalize).vNE()]
        return(minusOneVNE)

    def minusOneResult(self):
        """
        Systematically remove one feature at a time and calculate the von Neumann entropy
        for the remaining data matrix. Allow multiprocessing.
        """
        self.feature_num = len(self.feature_names)
        result_summary = [[] for i in range(self.feature_num)]

        with Pool(processes = self.num_cpu) as p:
            with tqdm(total = self.feature_num) as pbar:
                for i, res in enumerate(p.imap_unordered(self.minusOneVNE, range(self.feature_num))):
                    result_summary[i] = res
                    pbar.update()

        self.result_summary = pandas.DataFrame(data = result_summary, columns = ['feature_no', 'feature_name', 'vNE'])

        return(self.result_summary)


class InfoRichCalling:
    """
    The difference between between this and MinusOneVNE class is that this class take the
    upper and the lower boundaries for information-rich feature choosing. MinusOneVNE class
    only focus on calculating the von Neumann entropy after systematically removal of column,
    not choosing feature.

    I envision in the future I might write different ways of choosing information-rich features,
    so I create this class to store all relevent information-rich feature calling mode if the self.data
    remains unchange.
    """

    def __init__(self, data, current_feature_names, upper_threshold_factor, lower_threshold_factor,
                 num_cpu, direct_from_result_summary, normalize, silence = False, suppress = False):
        if len(direct_from_result_summary) == 0:
            self.data = numpy.array(data)
            self.feature_names = current_feature_names
            self.normalize = normalize
            self.current_result_summary = MinusOneVNE(data = self.data, normalize = self.normalize, feature_names = self.feature_names, num_cpu = num_cpu).minusOneResult()
            self.start_from_data = True
            self.force_output = False
        else:
            #self.data = data                           # not important when set direct_from_result_summary = True. Can even be ''.
            #self.feature_names = current_feature_names # not important when set direct_from_result_summary = True. Can even be ''.
            self.current_result_summary = direct_from_result_summary
            self.start_from_data = False
            self.force_output = True                    # avoid abort the program when arise ErrorCode4

        self.upper_threshold_factor = upper_threshold_factor
        self.lower_threshold_factor = lower_threshold_factor
        self.suppress = suppress
        self.silence = silence
        self.warning_code = ''


    def infoRich(self):
        """
        Select information-rich features based on the thresholds and the von Neumnann entropy.
        """

        # get threshold
        entro_mean = numpy.mean(numpy.array(self.current_result_summary["vNE"]))
        entro_sd = numpy.std(numpy.array(self.current_result_summary["vNE"]), ddof = 1)

        if self.upper_threshold_factor != 'None':
            self.upper_threshold = entro_mean + (self.upper_threshold_factor * entro_sd)
        if self.lower_threshold_factor != 'None':
            self.lower_threshold = entro_mean - (self.lower_threshold_factor * entro_sd)

        if self.upper_threshold_factor != 'None' and self.lower_threshold_factor != 'None':
            pass_condition = [(self.current_result_summary["vNE"] > self.upper_threshold) | (self.current_result_summary["vNE"] < self.lower_threshold)]
        elif self.upper_threshold_factor != 'None':
            pass_condition = [(self.current_result_summary["vNE"] > self.upper_threshold)]
        elif self.lower_threshold_factor != 'None':
            pass_condition = [(self.current_result_summary["vNE"] < self.lower_threshold)]

        tag = [1]
        self.current_result_summary["info_rich_feature"] = numpy.select(pass_condition, tag, default = 0)
        self.result_summary = self.current_result_summary

        info_rich_num = sum(numpy.array(self.current_result_summary["info_rich_feature"]))

        try:
            if info_rich_num == 0 and self.force_output == False:
                raise Error(code = 4)
        except Error as e:
            raise ErrorCode4(suppress = self.suppress) from e

        # maximum number of information-rich features
        if self.start_from_data == True:
            max_num = numpy.minimum(numpy.size(self.data, 0), numpy.size(self.data, 1))
            self.current_result_summary["max_num_for_this_dataset"] = [max_num] * self.current_result_summary.shape[0]
        else:
            max_num = self.current_result_summary["max_num_for_this_dataset"].iloc[0]

        try:
            if info_rich_num > max_num:
                raise WarningCode11(silence = self.silence)

        except WarningCode11:
            self.warning_code = '11'
            self.result_summary['deviate_from_mean'] = self.result_summary['vNE'] - entro_mean

            if self.upper_threshold_factor != 'None':
                if self.lower_threshold_factor != 'None':
                    self.result_summary['deviate_from_mean'] = abs(self.result_summary['deviate_from_mean'] - entro_mean)
                elif self.lower_threshold_factor == 'None':
                    self.result_summary['deviate_from_mean'] = self.result_summary['deviate_from_mean'] - entro_mean # TODO: need test

            elif self.upper_threshold_factor != 'None' and self.lower_threshold_factor != 'None':
                self.result_summary['deviate_from_mean'] = -(self.result_summary['deviate_from_mean'] - entro_mean)  # TODO: need test

            # rank feature-rich feature and select the ones that have higher impact on vNE when removal
            self.result_summary.sort_values(by = 'deviate_from_mean', ascending= False)
            self.reselect = list(self.result_summary[self.result_summary['info_rich_feature'] == 1].iloc[0:max_num]['feature_name'])  ##TODO: test this!!

            reselect_TF = [self.result_summary['feature_name'].isin(self.reselect)]
            self.result_summary["before_reselection"] = self.result_summary["info_rich_feature"]
            self.result_summary["info_rich_feature"] = numpy.select(reselect_TF, [1], default = 0)


    def infoRichSelect(self):
        """
        Subset infoRich result and only retain that corresponding to information-rich features
        """
        self.infoRich()
        self.info_rich_feature = self.result_summary.loc[self.result_summary["info_rich_feature"] > 0]   # <class 'pandas.core.frame.DataFrame'>


def reproducibility(InfoRichCalling_class, infoRich_dict, nrow, basename, vNE_output_folder,
                    output_file_tag, direct_from_result_summary, num_cpu, normalize):
    ## Need to be careful about the data and feature_names. They should be updated if user
    ## call the filtering function. To avoid confusion, I decided to not to list this
    ## function under MinusOneVNE or InfoRichCalling
    """
    Calculate the reproducibility of information-rich feature calling.

    This function and reproducibility_summary() are separated from InfoRichCalling_class because
    1. Aovid potential confusion regrading the data matrix (-1 row compared to the data matrix in
       InfoRichCalling_class)
    2. Easier for multiprocessing
    """
    for j in range(nrow):
        # delete row by index
        jackknift_subset = numpy.delete(InfoRichCalling_class.data, j, 0)

        info_rich_result = InfoRichCalling(data = jackknift_subset,
                                           current_feature_names = InfoRichCalling_class.feature_names,
                                           upper_threshold_factor = InfoRichCalling_class.upper_threshold_factor,
                                           lower_threshold_factor = InfoRichCalling_class.lower_threshold_factor,
                                           num_cpu = num_cpu, normalize = normalize,
                                           direct_from_result_summary = direct_from_result_summary)

        info_rich_result.infoRichSelect()

        detail_calling_result = info_rich_result.result_summary
        detail_calling_result['info_rich_feature'] = detail_calling_result['info_rich_feature'].replace([0, 1], ["No", "Yes"])
        detail_vNE_file_name = os.path.join(vNE_output_folder, basename.replace(".csv", "")) + output_file_tag + "__sub_" + str(j) + "_detail_vNE.csv"
        detail_calling_result.to_csv(detail_vNE_file_name)

        for element in info_rich_result.info_rich_feature['feature_name']:
            if element in infoRich_dict.keys():
                infoRich_dict[element] += 1
            else:
                infoRich_dict[element] = 1
    return(infoRich_dict)


def reproducibility_summary(filtered_matrix, infoRich_dict):
    filtered_matrix = numpy.array(filtered_matrix)
    nrow = numpy.size(filtered_matrix, 0)

    infoRich_dict_to_list = []

    for key, value in infoRich_dict.items():
        infoRich_dict_to_list_sub = [key, value]
        infoRich_dict_to_list.append(infoRich_dict_to_list_sub)

    infoRich_reproducibility = pandas.DataFrame(data = infoRich_dict_to_list,
                                                columns=['feature_name', 'occurrence'])
    infoRich_reproducibility['repreducibility (%)'] = infoRich_reproducibility['occurrence'] / nrow * 100
    return(infoRich_reproducibility)


class Kernal:
    """
    Smallest funcitonal unit
    """

    def __init__(self, file_name, detection_limit, tolerance, filter, upper_lim, lower_lim,
                 infoRich_threshold, quick_look, use_fractional_abundance, vNE_output_folder,
                 output_file_tag, num_cpu, notebook_name, normalize, neglect = False, silence = False,
                 suppress = False):

        self.input_matrix = RawDataImport(file_name = file_name, for_merging_file = False,
                                          suppress = False, second_chance = False)
        self.basename = os.path.basename(file_name)
        print('\nworking on: ' + self.basename)

        self.input_matrix.readCSV()
        self.detection_limit = detection_limit
        self.tolerance = tolerance
        self.filter = filter
        self.upper_lim = upper_lim
        self.lower_lim = lower_lim
        self.infoRich_threshold = infoRich_threshold
        self.quick_look = quick_look
        self.use_fractional_abundance = use_fractional_abundance
        self.vNE_output_folder = vNE_output_folder
        self.output_file_tag = output_file_tag
        self.notebook_name = notebook_name
        self.num_cpu = num_cpu
        self.neglect = neglect
        self.normalize = normalize
        self.silence = silence
        self.suppress = suppress


    def importAndProcess(self):
        if self.use_fractional_abundance == True:
            self.input_matrix.relativeAbundance()
            self.input_matrix.raw_data_before_filter = self.input_matrix.data
        else:
            self.input_matrix.raw_data_before_filter = self.input_matrix.raw_data

        if self.detection_limit > 0:
            try:
                if max(self.input_matrix.data.max()) < self.detection_limit:
                    raise Error(code = 6)
            except Error as e:
                raise ErrorCode6(suppress = self.suppress)

            self.input_matrix.detectionLimit(detection_limit = self.detection_limit)

        if self.filter:
          if self.filter == 'HardFilter':
              self.input_matrix.hardFilter(zero_tolerance_level = self.tolerance)
              self.filtered_data = self.input_matrix
          else:
              self.filtered_data = self.input_matrix

        num_sample = len(self.input_matrix.sample_id)
        num_feature = len(self.input_matrix.feature_names)
        num_feature_removed = self.input_matrix.raw_data.shape[1] - self.filtered_data.data.shape[1]

        notebook = UpdateNoteBook(notebook_name = self.notebook_name, neglect = self.neglect)
        notebook.updateFilterResult(num_sample = num_sample, num_feature = num_feature, num_feature_removed = num_feature_removed, basename = self.basename)


    def infoRichCallingAndReproducibility(self):
        self.info_rich_result = InfoRichCalling(data = self.filtered_data.data, current_feature_names = self.filtered_data.feature_names,
                                                upper_threshold_factor = self.upper_lim, lower_threshold_factor = self.lower_lim,
                                                num_cpu = self.num_cpu, normalize = self.normalize, direct_from_result_summary = '', silence = self.silence)

        if self.quick_look == True:
            print("Feature reduction with emmer...")

            self.info_rich_result.infoRichSelect()
            self.warning_code = self.info_rich_result.warning_code

            self.detail_calling_result = self.info_rich_result.result_summary
            self.detail_calling_result['info_rich_feature'] = self.detail_calling_result['info_rich_feature'].replace([0, 1], ["No", "Yes"])
            self.list_of_info_rich_features = list(self.info_rich_result.info_rich_feature['feature_name'])

            detail_vNE_file_name = os.path.join(self.vNE_output_folder, self.basename.replace(".csv", "")) + self.output_file_tag + "__detail_vNE.csv"
            self.detail_calling_result.to_csv(detail_vNE_file_name)

        else:
            print("\nCalculating the reproducibility of information-rich feature calling...")
            self.infoRich_dict = {}
            self.nrow = numpy.size(numpy.array(self.info_rich_result.data), 0)

            infoRich_dict = reproducibility(InfoRichCalling_class = self.info_rich_result, infoRich_dict = self.infoRich_dict,
                                            nrow = self.nrow, basename = self.basename, vNE_output_folder = self.vNE_output_folder,
                                            output_file_tag = self.output_file_tag, normalize = self.normalize, num_cpu = self.num_cpu,
                                            direct_from_result_summary = '')
            infoRich_dict_filtered = {key:val for key, val in infoRich_dict.items() if val >= self.infoRich_threshold}

            info_rich_features_w_reproducibility = reproducibility_summary(self.info_rich_result.data, infoRich_dict_filtered)

            r, c = info_rich_features_w_reproducibility.shape
            max_num = numpy.minimum(numpy.size(self.info_rich_result.data, 0), numpy.size(self.info_rich_result.data, 1))

            try:
                if r > max_num:
                    raise WarningCode12(silence = self.silence)
                else:
                    self.info_rich_features_w_reproducibility = info_rich_features_w_reproducibility
            except WarningCode12:
                self.warning_code = '12'
                info_rich_features_w_reproducibility = info_rich_features_w_reproducibility.sort_values(by = 'occurrence', ascending = False)
                self.info_rich_features_w_reproducibility = info_rich_features_w_reproducibility.iloc[0:max_num]

            self.list_of_info_rich_features = list(self.info_rich_features_w_reproducibility['feature_name'])

        notebook = UpdateNoteBook(notebook_name = self.notebook_name, neglect = self.neglect)
        notebook.updateEmmerResult(num_info_rich = len(self.list_of_info_rich_features))



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_harvest

from ..main.basic.math import NonDesityMatrix
from ..main.basic.read import RawDataImport, GetFiles
from ..main.advanced.iteration import MinusOneVNE, InfoRichCalling, reproducibility, reproducibility_summary, Kernal
#from ..main.advanced.iteration import MinusOneVNE, MinDataLostFilter, InfoRichCalling, reproducibility, reproducibility_summary, Kernal
from ..harvest import HarvestArgs, EMMER, mergeDataFrame
from ..troubleshoot.err.error import *

from pandas.util.testing import assert_frame_equal

import unittest
import argparse
import shutil
import pandas
import numpy
import glob
import sys
import os


class TestHarvestArgs(unittest.TestCase):

    def test_getArgsI(self):
        print('\ntest_TestHarvestArgs.getArgsI:')
        print('        case 1: non csv file when input a specific file')
        sys.argv[1:] = ['-i', 'emmer/data/sow_test_dir_2/targert_file_1.txt']
        with self.assertRaises(ErrorCode1):
            processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
            processed_args.getArgsI()
        print('===========================================================')


    def test_getArgsQT(self):
        print('\ntest_TestHarvestArgs.getArgsQT:')
        print('        case 1: do not need to set args.t when use args.q')
        sys.argv[1:] = ['-t', '2', '-q']
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsQT()
        my_result = processed_args.warning_code
        expected_result = '10'
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_getArgsFZ(self):
        print('\ntest_TestHarvestArgs.getArgsFZ:')
        print('        case 1: test error handling')
        print('             1.1: unexpected agrs.z number setting when using HardFilter')
        sys.argv[1:] = ['-f', 'HardFilter', '-z', '2']
        with self.assertRaises(ErrorCode2):
            processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
            processed_args.getArgsFZ()

        print('        ---------------------------------------------------')
        print('             1.2: missing agrs.z when using HardFilter')
        sys.argv[1:] = ['-f', 'HardFilter']
        with self.assertRaises(ErrorCode3):
            processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
            processed_args.getArgsFZ()

        print('        ---------------------------------------------------')
        print('             1.3: set agrs.z when using None filter')
        sys.argv[1:] = ['-f', 'None', '-z', '0.5']
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsFZ()
        my_result = processed_args.warning_code
        expected_result = '2'
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_getArgsUL(self):
        print('\ntest_TestHarvestArgs.getArgsUL:')
        print('        case 1: missing both args.u and args.l settings')
        sys.argv[1:] = ['-f', 'None', '-z', '0.5']
        with self.assertRaises(ErrorCode5):
            processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
            processed_args.getArgsUL()
        print('===========================================================')


    def test_getArgsPS(self):
        print('\ntest_TestHarvestArgs.getArgsPS:')
        print('        case 1: args.s warning handling')
        print('             1.1: current version of emmer can not generate args.s plots when working on specific csv (args.i)')
        sys.argv[1:] = ['-i', 'emmer/data/data_dir_3/group_A.csv', '-s']
        #with self.assertRaises(WarningCode3):
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsI()
        processed_args.getArgsPS()
        my_result = processed_args.warning_code
        expected_result = '3'
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             1.2: current version of emmer can not generate args.s plots when input directory only contains one csv file')
        sys.argv[1:] = ['-i', 'emmer/data/data_dir_1/', '-s']
        #with self.assertRaises(WarningCode3):
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsI()
        processed_args.getArgsPS()
        my_result = processed_args.warning_code
        expected_result = '3'
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        case 2: args.p warning handling')
        print('             2.1: current version of emmer can not generate args.p plots when working on specific csv (args.i)')
        sys.argv[1:] = ['-i', 'emmer/data/data_dir_3/group_A.csv', '-p']
        #with self.assertRaises(WarningCode5):
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsI()
        processed_args.getArgsPS()
        my_result = processed_args.warning_code
        expected_result = '5'
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             2.2: current version of emmer can not generate args.p plots when input directory only contains one csv file')
        sys.argv[1:] = ['-i', 'emmer/data/data_dir_1', '-p']
        #with self.assertRaises(WarningCode6):
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsI()
        processed_args.getArgsPS()
        my_result = processed_args.warning_code
        expected_result = '5'
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_getArgsC(self):
        print('\ntest_TestHarvestArgs.getArgsC:')
        print('        case 1: args.c warning handling')
        print('             1.1: user set args.c at 0')
        sys.argv[1:] = ['-c', '0']
        with self.assertRaises(ErrorCode47):
            processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
            processed_args.getArgsC()

        print('        ---------------------------------------------------')
        print('             1.2: args.c > CPU in the computer')
        sys.argv[1:] = ['-c', '10000000000']
        with self.assertRaises(ErrorCode47):
            processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
            processed_args.getArgsC()

        print('        ---------------------------------------------------')
        print('        case 2: default setting')
        sys.argv[1:] = []
        processed_args = HarvestArgs(suppress = True, silence = False, neglect = True)
        processed_args.getArgsC()
        my_result = processed_args.num_cpu
        expected_result = 1
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestEMMER(unittest.TestCase):

    def test_EMMER(self):
        print('\ntest_EMMER:')
        print('    input_dir: "emmer/data/data_dir_1"')
        input_dir = 'emmer/data/data_dir_1'
        print('    output_file_tag: "test1"')
        output_file_tag = 'test1'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: False')
        specific_csv = False
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look: True')
        quick_look = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)
        my_result = one_file.input_file_names

        expected_result = ['emmer/data/data_dir_1/test_case_1.csv']
        self.assertListEqual(my_result, expected_result)
        shutil.rmtree('output')
        print('===========================================================')


    def test_singleFile(self):
        print('        ---------------------------------------------------')
        print('        case 1: HardFilter')
        print('             1.1: hypothetical data')
        print('    input_dir: "emmer/data/data_dir_1"')
        input_dir = 'emmer/data/data_dir_1'
        print('    output_file_tag: "test1"')
        output_file_tag = 'test1'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 0.6')
        tolerance = 0.6
        print('    filter: "HardFilter"')
        filter = 'HardFilter'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: True')
        specific_csv = False
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look: True')
        quick_look = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)

        one_file.singleFile()

        my_result = list(one_file.data.filtered_data.data.columns.values)

        expected_result = ['col1', 'col2', 'col3', 'col5', 'col6']
        self.assertListEqual(my_result, expected_result)

        ## filter: HardFilter; real data
        print('        ---------------------------------------------------')
        print('             1.2: read data (check each filtering steps)')
        print('             1.2.1: raw data')
        print('    input_dir: "emmer/data/data_dir_3/group_A.csv"')
        input_dir = 'emmer/data/data_dir_3/group_A.csv'
        print('    output_file_tag: "test1"')
        output_file_tag = 'test1'
        print('    detection_limit: 0.001')
        detection_limit = 0.001
        print('    tolerance: 0.33')
        tolerance = 0.33
        print('    filter: "HardFilter"')
        filter = 'HardFilter'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: True')
        specific_csv = True
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look: True')
        quick_look = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)

        one_file.singleFile()


        # Raw data; after removing empty rows and columns
        my_result = list(one_file.data.input_matrix.raw_data.shape)
        expected_result = [13, 4809]
        self.assertListEqual(my_result, expected_result)


        # After removing empty rows and columns
        print('        ---------------------------------------------------')
        print('             1.2.2: after removing empty rows and columns')
        my_result = list(one_file.data.input_matrix.raw_data_before_filter.shape)
        expected_result = [13, 1077]
        self.assertListEqual(my_result, expected_result)


        # After data filtering
        print('        ---------------------------------------------------')
        print('             1.2.3: after filtering')
        my_result = list(one_file.data.input_matrix.data.shape)
        expected_result = [13, 126]
        self.assertListEqual(my_result, expected_result)


        ## filter: None
        print('        ---------------------------------------------------')
        print('        case 2: No filter')
        print('    input_dir: "emmer/data/data_dir_1"')
        input_dir = 'emmer/data/data_dir_1'
        print('    output_file_tag: "test1"')
        output_file_tag = 'test1'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: False')
        specific_csv = False
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look: True')
        quick_look = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)

        one_file.singleFile()
        my_result = list(one_file.data.filtered_data.data.columns.values)

        expected_result = ['col1', 'col2', 'col3', 'col4', 'col5', 'col6']
        self.assertListEqual(my_result, expected_result)
        shutil.rmtree('output')
        print('===========================================================')


    def test_multipleFiles(self):
        print('        ---------------------------------------------------')
        print('        case 1: set quick_look at False')
        print('    input_dir: "emmer/data/data_dir_2"')
        input_dir = 'emmer/data/data_dir_2'
        print('    output_file_tag: "multipleFiles_test1"')
        output_file_tag = 'multipleFiles_test2'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: False')
        specific_csv = False
        print('    information-rich threshold: 2')
        infoRich_threshold = 2
        print('    quick_look: False')
        quick_look = False
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)

        one_file.multipleFiles()

        my_result = one_file.summary_df

        data = [[0.00, 28.57],
                [50.00, 0.00],
                [66.67, 0.00],
                [0.00, 100.00],
                [33.33, 0.00],
                [66.67, 0.00]]
        expected_result = pandas.DataFrame(data, columns = ['test_case_1.csv', 'test_case_2.csv'],
                                           index = ['col1', 'col2', 'col3', 'col4', 'col5', 'col6'])

        assert_frame_equal(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        case 2: set quick_look at True')
        print('    input_dir: "emmer/data/data_dir_2"')
        input_dir = 'emmer/data/data_dir_2'
        print('    output_file_tag: "multipleFiles_test1"')
        output_file_tag = 'multipleFiles_test2'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: False')
        specific_csv = False
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look: True')
        quick_look = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)

        one_file.multipleFiles()

        my_result = one_file.summary_df

        data = [[1.0, 0.0],
                [1.0, 0.0],
                [0.0, 1.0],
                [1.0, 0.0]]
        expected_result = pandas.DataFrame(data, columns = ['test_case_1.csv', 'test_case_2.csv'],
                                           index = ['col2', 'col3', 'col4', 'col6'])

        assert_frame_equal(my_result, expected_result)
        shutil.rmtree('output')
        print('===========================================================')


class TestMergeDataFrame(unittest.TestCase):
    def test_mergeDataFrame(self):
        print('\ntest_mergeDataFrame:')
        print('    input_dir: "emmer/data/data_dir_2"')
        input_dir = 'emmer/data/data_dir_2'
        print('    output_file_tag: "test2"')
        output_file_tag = 'test2'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_threshold_factor: 1')
        upper_threshold_factor = 1
        print('    lower_threshold_factor: 1')
        lower_threshold_factor = 1
        print('    specific_csv: False')
        specific_csv = False
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look: True')
        quick_look = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        one_file = EMMER(input_dir = input_dir, output_file_tag = output_file_tag,
                         detection_limit = detection_limit, tolerance = tolerance,
                         filter = filter, upper_threshold_factor = upper_threshold_factor,
                         lower_threshold_factor = lower_threshold_factor,
                         specific_csv = specific_csv, infoRich_threshold = infoRich_threshold,
                         notebook_name = '', neglect = '', normalize = normalize,
                         num_cpu = num_cpu, quick_look = quick_look,
                         use_fractional_abundance = use_fractional_abundance)

        one_file.multipleFiles()
        transform_info = mergeDataFrame(EMMER_class = one_file, select = 'filtered_infoRich',
                                        file_name_list = one_file.clean_df_file_names,
                                        info_rich_list = one_file.collections_of_info_rich_features,
                                        notebook_name = '', normalize = normalize, neglect = True)

        my_result = one_file.merged_dataframe.shape
        expected_result = (13, 4)

        self.assertEqual(my_result, expected_result)
        shutil.rmtree('output')
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#from support import printForUnittest
import sys
import io

__version__ = 'v0.8.c.4+'

"""
=========================================================================================
Handle all the warning messages in EMMER.

In EMMER, errors always need to be address to ensure the completely excution. Warnings, on
the other hand, are usually non-critical and can be automatically fixed by EMMER.
==========================================================================================
"""

#class Warning(Exception):  # TODO: to retire (maybe)
#    """
#    Warning class
#
#    Arguments:
#        code -- Type: str
#                error code number in str
#
#    Attributes:
#        code -- Type: str
#    """
#    def __init__(self, code):
#        self.code = code


def reportWarning(fn):
    """
    What should emmer behave when warning occurs?
    1. Whether to print the docstring for WarningCode __init__ function.
       It is much easier to write detail explanation for each error or
       warning in docstring than in print()
    """
    def wapper(*args, **kwargs):

        silence = fn(*args, **kwargs)
        if silence == False:
            print(fn.__doc__)

        return

    return wapper


class WarningCode1(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 1]]
        Noncritial parameter setting error:
        When setting filter as "MinDataLostFilter", you do not have to set -z.

        emmer.harvest will neglect the -z argument and keep running.
        """
        return(silence)


class WarningCode2(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 2]]
        Noncritial parameter setting error:
        When setting filter as "None", you do not have to set -z.

        emmer.harvest will neglect the -z argument and keep running.
        """
        return(silence)


class WarningCode3(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 3]]
        Noncritial parameter setting error:
        Currently emmer does not support -sanityCheck (-s) option when designate a specific
        csv file or when designate a directory that only store one csv file.

        emmer.harvest will reset sanityCheck to False and keep running.
        """
        return(silence)

## TODO:
#  WarningCode4 available ###############################################################################################
## TODO


class WarningCode5(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 5]]
        Noncritial parameter setting error:
        Currently emmer does not support -plot (-p) option when designate a specific
        csv file or when designate a directory that only store one csv file.

        emmer.harvest will reset -plot to False and keep running.
        """
        return(silence)


class WarningCode6(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 6]]
        'output' folder exist in your working directory. Excuting emmer.tests will
        generate additional files in 'output' folder. Please consider to move your
        existing output file to other directory. Or remove 'output' folder to avoid
        confusion in the feature.
        """
        return(silence)


class WarningCode7(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 7]]
        Noncritial parameter setting error:
        Missing args.b (-b) setting. Will set the number of bin for generating
        histogram as 20
        """


class WarningCode8(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 8]]
        We are deeply sorry that the current version of emmer does not allow users
        to change shapes of data points on the PCA plot.

        This limitation might be related to issue 11155
        (https://github.com/matplotlib/matplotlib/issues/11155)

        We will try to enable the functions that allow user to define shapes of
        data points once this issue is resolved.

        If you have better solution, please contact emmer deveplors
        """
        return(silence)


class WarningCode9(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 9]]
        Expect arguments for -m
        """
        return(silence)


class WarningCode10(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 10]]
        Noncritial parameter setting error:
        You do not need to set threshold for reporting information-rich feature
        (-t) when using Quick Look Mode (-q).

        emmer.harvest will neglect the -t setting and keep running.
        """
        return(silence)


class WarningCode11(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 11]]
        Number of information-rich feature should not exceed the minimum number
        between number of row and number of column.
        """
        return(silence)


class WarningCode12(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 12]]
        Number of information-rich feature should not exceed the minimum number
        between number of row and number of column. Remove feature with less
        reproducibility before reprot the final list of information-rich features.
        """
        return(silence)


class WarningCode13(Warning):

    @reportWarning
    def __init__(self, silence):
        """
        [[Warning code 13]]
        Redundant and maybe conflicting parameter setting:
        When choose -m as ['Permanova', 'Individual'] and provide provide -p,
        emmer.bake will use the corrodinates in -p (not -i).

        To avoid potentially conflicting information regrading the coordinates, emmer.bake
        will neglect -i setting and use the coordinates from -p.
        """
        return(silence)



#!/usr/bin/env python3


from .posthoc.stats.revisit_thresholds import revisitThresholdResult
from .posthoc.stats.bifurication import identifiedFeatures
from .posthoc.stats.reproducibility import reproSummary
from .posthoc.stats.permanova import permanovaResult

from .posthoc.visual.projection import ProjectionArgs, projectNew
from .posthoc.visual.individual import plotIndividual

from .toolbox.recorder import initNoteBook, UpdateNoteBook

from .troubleshoot.warn.warning import WarningCode9
from .troubleshoot.inquire.input import InputCode3

import argparse
import sys
import os

__version__ = '1.0'

"""
This module (emmer.bake) take in EMMER output csv files and generate plots or conduct
follow-up analyses.
"""

##==0==##
def tutorial(self):
    """
    ##==1==## Different Modes and Their Usages
    1. Individual: plot individual samples onto the PCA plot. Allow better flexibility in
                   aesthetics setting
    2. Permanova: apply PERMANOVA test on different groups in the PCA plot
    3. RevisitThreshold: revisit threshold settings (args.t, args.l, and args.u) used in
                         emmer.harvest
    4. Reproducibility: summarizing information-rich feature calling reproducibility into
                        basic statistics and histogram
    5. Bifurication: identify information-rich features that help to differentiate different
                     groups on the PCA space.
    6. Projection: project new observations onto the existing PCA space


    ##==2==##  How To
    0. General options:
       -m M, -mode        Which mode you want to use. Please refer to the Different Modes and Their Usages section above.

       optional:
       -o O               Expect a string. This string will be tag to all your output file name from the analysis.
       optional:
       -w W, -writeDownDetails
                          Do you want to add additional notes as emmer run? Default: False.
                          (Not available,. Will be included in future update)
                          Usage:
                          python3 -m emmer.harvest <other_arguments_and_inputs> -w
    ------------------------------------------------------------
    1. Individual mode:
       python3 -m emmer.bake -m 'Individual' -i emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv

       -i I               An emmer.harvest-gereated projection file (whose file name contains the keyword 'projection' or
                          'coordinates') that stores the coordinates of each sample in PCA space.
       -a A               An emmer.harvest-gereated file that contains information about percent explained by each PC. This
                          file cound be found under 'output' with 'precent_explained' in the file name.

       optional:
       -p P               An csv file with PCA coordinates and aesthetics settings. This file will be generated after
                          you run Individual mode. You can find the file under 'revisit/''
                          You do not have to provide -i argument if you are using -p option. When providing arguments for
                          both -i and -p, emmer will ignore -i input and report warning message.

                          Should you choose to rerun this mode with, you can use
                          python3 -m emmer.bake -m 'Individual' -p where_you_store_the_file/coordinates_and_parameters.csv
    ------------------------------------------------------------
    2. Permanova mode:
       python3 -m emmer.bake -m 'Permanova' -i emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv

       -i I               An emmer.heavest-gereated projection file (whose file name contains the keyword 'projection' or
                          'coordinates') that stores the coordinates of each sample in PCA space.
    ------------------------------------------------------------
    3. Reproducibility mode:
       python3 -m emmer.bake -m 'Reproducibility' -b 20 -i emmer/data/bake_data_dir_4/information_rich_features_summary.csv

       -b B               Number of bins used when generating the histogram under "Reproducibility" mode
       -i I               An emmer.harvest-gereated csv file that stores information-rich feature calling reproducibility
    ------------------------------------------------------------
    4. RevisitThreshold mode:
       python3 -m emmer.bake -m 'RevisitThreshold' -u 2.5,1.5,0.25 -l 2.5,1.5,0.25 -t 2,2,0 -e output/detail_vNE/ -i output/filtered_data/

       -u U               Revisit args.u (-u) setting in EMMER. Expect a three-element tuple. Example: 3,1,1 (upper bondary,
                          lower bondary, increment)
       -l L               Revisit args.l (-l) setting in EMMER. Expect a three-element tuple. Example: 3,1,1 (upper bondary,
                          lower bondary, increment)
       -t T               Revisit args.t (-t) setting in EMMER. Expect a three-element tuple. Example: 3,1,1 (upper bondary,
                          lower bondary, increment)
       -e E               A directory that store detail_vNE.csv files. The input is most likely to be 'output/detal_vNE'
       -i I               A directory that store filtered or prefilter data. The input is most likely to be 'output/filtered_data' or
                          'output/pre_filter_data'
                          Choose the folder based on whether you want to PCA plot generated from the dataset that only contains information-
                          rich features to be similar to the PCA generated from filtered_data or pre_filter_data

       optional:
       -n N, normalize    Choose this option when you set -n as True when running emmer.harvest
       -c C, -cpuNum      Support multiprocessing. This argument is used to set the number of CPU used in the analysis.
    ------------------------------------------------------------
    5. Bifurication mode:
       python3 -m emmer.bake -m 'Bifurication' -i emmer/data/bake_data_dir_4/filtered_data
                             -p emmer/data/bake_data_dir_4/information_rich_features_summary.csv
       [!] Remember to covert the above command into a single-line command if you wish to run the example

       -i I               A directory that store filtered data. The input is most likely to be 'output/filtered_data'
       -p P               An emmer-generated summary of the information-rich features.

       optional:
       -n N, normalize    Choose this option when you set -n as True when running emmer.harvest
    ------------------------------------------------------------
    6. Projection mode:
       python3 -m emmer.bake -m 'Projection' -i emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv
                             -v emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv
                             -s emmer/data/bake_data_dir_9/filtered_infoRich__data_scaler.csv
                             -x emmer/data/bake_data_dir_9/new_observation.csv -r
       [!] Remember to covert the above command into a single-line command if you wish to run the example

       -i I                An emmer.harvest-gereated projection file (whose file name contains the keyword 'projection' or
                           'coordinates') that stores the coordinates of each sample in PCA space.
       -v V                An emmer.harvest generated csv file that stores the transformation matix. Hint: the file name that
                           contains 'transformation_matrix'.
       -s1 S1              An emmer.harvest generated csv file that stores the column means. Hint: the file name that contains '__data_colmean'
       -x X                New observations. Input directory that contains one or many csv files or a path to specific csv file.
                           Expect to have column headers and row names. Each row represents a sample and each column represents a feature.

       optional:
       -r R                Convert data into fractional abudance.
       -s2 S2              An emmer.harvest generated csv file that stores the standard deviation for each column. EMMER will generate this
                           csv file, when you set -n as True when running emmer.harvest.
                           Hint: the file name that contains '__data_colstd'
    """
    pass


class BakeCommonArgs:
    """
    Take common arguments for different modes when running bake modules

    Objective: so we can test and use @

    Argument:
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        silence -- Type: boolean
                   emmer will not report warning massage when silence == True
        neglect -- Type: boolean
                   Suppress notebook initation when running unittest
        test -- Type: boolean
                Avoid running InputCode3 after trigger WarningCode9 during unittest

    Attributes:
        suppress -- Type: boolean
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        output_file_tag -- Type: str
                           Corresponding to args.o
        selected_model -- Type: str
                          Corresponding to args.m
        warning_code -- Type: str
                        For unittest
    """

    def __init__(self, suppress, silence, neglect, test):
        parser = argparse.ArgumentParser(description = '#############################################################################\nPlease use -g when you need additional explanation on different modes their corresponding arguments. Try: python3 -m emmer.harvest -g\n#############################################################################')
        parser.add_argument('-g', '-guide', action = 'store_true')
        parser.add_argument('-i', type = str)
        parser.add_argument('-a', type = str)
        parser.add_argument('-p', type = str)
        parser.add_argument('-o', type = str)
        parser.add_argument('-m', '-mode')
        parser.add_argument('-e', type = str)
        parser.add_argument('-n', '-normalize', action = 'store_true')
        parser.add_argument('-t', type = str)
        parser.add_argument('-u', type = str)
        parser.add_argument('-l', type = str)
        parser.add_argument('-b', type = int)
        parser.add_argument('-r', action = 'store_true')
        parser.add_argument('-v', type = str)
        parser.add_argument('-s1', type = str)
        parser.add_argument('-s2', type = str)
        parser.add_argument('-x', type = str)
        parser.add_argument('-c', '-cpuNum', type = int)
        parser.add_argument('-w', '-writeDownDetails', action = 'store_true')
        self.args = parser.parse_args()
        self.suppress = suppress
        self.neglect = neglect
        self.silence = silence
        self.test = test

        if self.args.g:
            print(tutorial.__doc__)
            sys.exit()


    def getArgsO(self):
        if self.args.o:
            self.output_file_tag = str(self.args.o)
        else:
            self.output_file_tag = ''


    def getArgsW(self):
        head, tail = os.path.split(os.getcwd())   # head: no last path component, which is 'revisit', in current working directory
        self.notebook_name = initNoteBook(current_wd = head, script_name = 'emmer.bake', script_version = __version__,
                                          tag = self.output_file_tag, neglect = self.neglect, explicit = self.args.w)

    def getArgsM(self):
        choice_dict = {'1': 'Individual',
                       '2': 'Permanova',
                       '3': 'RevisitThreshold',
                       '4': 'Reproducibility',
                       '5': 'Bifurication',
                       '6': 'Projection'}

        try:
            if self.args.m not in choice_dict.values() and self.args.m not in choice_dict.keys():
                raise WarningCode9(silence = self.silence)
            else:
                self.selected_model = self.args.m
        except WarningCode9:
            self.warning_code = '9'
            if self.test == False:
                model = InputCode3(set = ['mode'], choice_dict = choice_dict, suppress = False)
                self.selected_model = model.selection


    def getHomeKeepingArgs(self):
        self.getArgsO()
        self.getArgsW()
        self.getArgsM()
        UpdateNoteBook(notebook_name = self.notebook_name, neglect = self.neglect).updateArgs(args = self.args)


##==1==## set input parameters
if __name__ == '__main__':
    current_wd = os.getcwd()
    retrospect_dir = os.path.join(current_wd, 'revisit')
    if not os.path.exists(retrospect_dir):
        os.makedirs(retrospect_dir)

    os.chdir(retrospect_dir)

    common_args = BakeCommonArgs(suppress = False, silence = False, neglect = False, test = False)
    common_args.getHomeKeepingArgs()


##==2==## different bake modes
##--1--## model = 'Individual'
    # python3 -m emmer.bake -m 'Individual' -i emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv
    if common_args.selected_model == 'Individual':
        plotIndividual(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
                       output_file_tag = common_args.output_file_tag, suppress = False, silence = False)


##--2--## model = 'Permanova'
    # python3 -m emmer.bake -m 'Permanova' -i emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv
    if common_args.selected_model == 'Permanova':
        permanovaResult(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
                        output_file_tag = common_args.output_file_tag, notebook_name = common_args.notebook_name,
                        suppress = False, silence = False, neglect = False)
#        permanovaResult(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
#                        output_file_tag = common_args.output_file_tag, suppress = False, silence = False)



##--3--## style = 'Reproducibility'
    # python3 -m emmer.bake -m 'Reproducibility' -b 20 -i emmer/data/bake_data_dir_4/information_rich_features_summary.csv
    if common_args.selected_model == 'Reproducibility':
        reproSummary(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
                     output_file_tag = common_args.output_file_tag, suppress = False, silence = False)


##--4--## model = 'RevisitThreshold'
    # python3 -m emmer.bake -m 'RevisitThreshold' -u 2.5,1.5,0.25 -l 2.5,1.5,0.25 -e output/detail_vNE/ -i output/filtered_data/
    if common_args.selected_model == 'RevisitThreshold':
        revisitThresholdResult(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
                               output_file_tag = common_args.output_file_tag, suppress = False, silence = False)


##--5--## model = 'Bifurication'
    # python3 -m emmer.bake -m 'Bifurication' -i emmer/data/bake_data_dir_4/filtered_data -p emmer/data/bake_data_dir_4/information_rich_features_summary.csv
    if common_args.selected_model == 'Bifurication':
        identifiedFeatures(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
                           output_file_tag = common_args.output_file_tag, suppress = False)


##--6--## model = 'Projection'
    # python3 -m emmer.bake -m 'Projection' -i emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv \
    #                       -v emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv \
    #                       -s emmer/data/bake_data_dir_9/filtered_infoRich__data_scaler.csv -x emmer/data/bake_data_dir_9/new_observation.csv -r
    if common_args.selected_model == 'Projection':
        new_and_old_projection_df = projectNew(args = common_args.args, current_wd = current_wd, retrospect_dir = retrospect_dir,
                                               output_file_tag = common_args.output_file_tag, suppress = False)



#!/usr/bin/env python3

from ...main.basic.math import NonDesityMatrix
from ...main.basic.read import RawDataImport
from ...toolbox.technical import emptyNumpyArray
from ...troubleshoot.warn.warning import WarningCode8

from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import pandas
import pylab
import numpy
import time
import sys
import os


class Projection:
    """
    Generate transformation matrix and scaler to project samples onto PCA space.
    Save those transformation matrix and scaler as attributes for easy access. Those
    attributes are important when projecting new observations onto the same PCA space.
    """

    def __init__(self, merged_dataframe, normalize):
        self.feature_names = list(merged_dataframe.columns.values)
        self.sample_id = list(merged_dataframe.index.values)
        self.data = numpy.array(merged_dataframe)
        self.df = merged_dataframe

        # get scaler (colmean)
        self.normalize = normalize
        self.non_density_matrix = NonDesityMatrix(self.data, normalize = self.normalize)

        # run normEigvals() so it will update colmean and matrixForSVD at the same time
        self.norm_eigvals = self.non_density_matrix.normEigvals()   # percent variance explained

        try:
            if len(self.norm_eigvals) < 2:
                raise Error(code = '11')
            elif len(self.norm_eigvals) == 2:
                self.plot_dimension = 2
            else:
                self.plot_dimension = 3
        except Error as e:
            if e.code == '11':
                raise ErrorCode11(suppress = False, second_chance = False) from e

        #self.scaler = self.non_density_matrix.colmean
        self.colmean = self.non_density_matrix.colmean
        self.epsilon = self.non_density_matrix.epsilon
        self.matrixForSVD = self.non_density_matrix.matrixForSVD

        u, s, vh = numpy.linalg.svd(self.matrixForSVD)

        self.Vh = vh
        self.V = numpy.transpose(vh[0:len(s),:])
        ## vh:
        #     row: principal components; col: features
        ## V:
        #     row: features; col: principal components

        self.U = u
        ## v:
        #     row: sample; col: principal components

        self.S = numpy.diag(s)
        ## s

        ncol_V = numpy.size(self.V, 1)
        self.PC_list = list(['PC' + str(i) for i in range(1 , (ncol_V + 1))])

        self.colmean_df = pandas.DataFrame([self.colmean], columns = self.feature_names, index = ['colmean'])
        self.V_df = pandas.DataFrame(self.V, columns = self.PC_list[0:len(s)], index = self.feature_names)

        nrow = numpy.size(self.data, 0)
        mean_centered_matrix = self.data - numpy.tile(self.colmean, (nrow, 1))

        if normalize == True:
            self.colstd = self.non_density_matrix.colstd
            self.colstd_df = pandas.DataFrame([self.colstd], columns = self.feature_names, index = ['colstd'])
            scaled_mean_centered_matrix = numpy.matmul(mean_centered_matrix, numpy.diag(1/self.colstd))
            projection = numpy.matmul(scaled_mean_centered_matrix, self.V)
        else:
            projection = numpy.matmul(mean_centered_matrix, self.V)

        self.projection_df = pandas.DataFrame(projection, columns = self.V_df.columns.values, index = self.sample_id)

        # for annotating PCA plot
        self.PC_annotation = pandas.DataFrame(data = emptyNumpyArray(nrow = len(self.norm_eigvals), ncol = 3),
                                              columns=['PC', 'percent explain', 'PC w percent explain'])
        self.PC_annotation['PC'] = self.PC_list
        self.PC_annotation['percent explain'] = list([numpy.round(i * 100, 2)for i in self.norm_eigvals])
        self.PC_annotation['PC w percent explain'] = self.PC_annotation["PC"].astype(str) + ' (' + self.PC_annotation['percent explain'].astype(str) + '%)'


    def cleanSpec(self):
        """
        If the input matrix can be decomposed by SVD into a spectrum of matrices, this function
        will clean the spectrum and only retain the matrices that corresponding to PC1, PC2, and,
        if available, PC3. Then this function will sum all of those matrices together and obtain
        a new matric that call self.spec_cleaned_data
        """
        S_select = pandas.DataFrame(self.S[:, 0:self.plot_dimension])
        S_zero = pandas.DataFrame(numpy.zeros([self.df.shape[0], (self.df.shape[1] - self.plot_dimension)]))
        S_adj = pandas.concat([S_select, S_zero], axis = 1, sort = True)
        self.S_adj = numpy.nan_to_num(S_adj)   # convert nan to 0

        U_times_S = numpy.matmul(self.U , self.S_adj)
        self.spec_cleaned_data = numpy.matmul(U_times_S , self.Vh)


class Plot:
    """
    Generate 2D or 3D PCA plot. Allow user to decide the best angle to present the PCA for best
    visualization experience.
    """

    def __init__(self, Projection_class_in_list, input_file_in_list_of_list, output_file_name, filter):#, PC_annotation_in_list):
        self.Projection_class_in_list = Projection_class_in_list
        self.input_file_in_list_of_list = input_file_in_list_of_list
        self.output_file_name = output_file_name
        self.filter = filter
        #self.PC_annotation_in_list = PC_annotation_in_list


    def extractData(self, current_dataset):
        # update self.current_dataset
        self.current_dataset = current_dataset

        # extract data
        self.current_coordinates = self.Projection_class_in_list[self.current_dataset].projection_df
        self.current_PC_list = self.Projection_class_in_list[self.current_dataset].PC_list
        self.current_PC_list_annotation = self.Projection_class_in_list[self.current_dataset].PC_annotation
        self.current_percent_explain = numpy.round((numpy.array(self.Projection_class_in_list[self.current_dataset].norm_eigvals) * 100), decimals = 2)

        self.current_dataset = 0 # reset iteration
        self.current_group = [os.path.basename(element).replace(".csv", "") for element in self.input_file_in_list_of_list[self.current_dataset]]


    def projectDots2D(self):
        """
        add dots on a 2D plot
        """
        for i in range(len(self.current_group)):
            self.projection_df_subset = self.current_coordinates[[self.current_group[i] in rowname for rowname in self.current_coordinates.index]]

            x = self.projection_df_subset['PC1']
            y = self.projection_df_subset['PC2']

            #add dots onto the plot
            self.ax.scatter(x, y, label = self.current_group[i], alpha = 1)

            self.ax.set_xlabel(self.current_PC_list[0] + ' (' + str(self.current_percent_explain[0]) + '%)')
            self.ax.set_ylabel(self.current_PC_list[1] + ' (' + str(self.current_percent_explain[1]) + '%)')

            self.ax.set_xlabel(self.current_PC_list_annotation['PC w percent explain'][0])
            self.ax.set_ylabel(self.current_PC_list_annotation['PC w percent explain'][1])


    def projectDots3D(self):
        """
        add dots on a 3D plot
        """
        for i in range(len(self.current_group)):
            self.projection_df_subset = self.current_coordinates[[self.current_group[i] in rowname for rowname in self.current_coordinates.index]]

            x = self.projection_df_subset['PC1']
            y = self.projection_df_subset['PC2']
            z = self.projection_df_subset['PC3']

            #add dots onto the plot
            self.ax.scatter(x, y, z, label = self.current_group[i], alpha = 1)   #

            self.ax.set_xlabel(self.current_PC_list[0] + ' (' + str(self.current_percent_explain[0]) + '%)')
            self.ax.set_ylabel(self.current_PC_list[1] + ' (' + str(self.current_percent_explain[1]) + '%)')
            self.ax.set_zlabel(self.current_PC_list[2] + ' (' + str(self.current_percent_explain[2]) + '%)')

            self.ax.set_xlabel(self.current_PC_list_annotation['PC w percent explain'][0])
            self.ax.set_ylabel(self.current_PC_list_annotation['PC w percent explain'][1])
            self.ax.set_zlabel(self.current_PC_list_annotation['PC w percent explain'][2])


    def assembleSubPlot(self, current_dataset, plot_location, title):
        self.extractData(current_dataset)

        try:
            if len(self.current_percent_explain) < 2:
                raise Error(code = '11')
            elif len(self.current_percent_explain) == 2:
                self.ax = self.fig.add_subplot(plot_location)
                self.projectDots2D()
            else:
                self.ax = self.fig.add_subplot(plot_location, projection='3d')
                self.projectDots3D()
        except Error as e:
            raise ErrorCode11(suppress = False) from e

        self.ax.legend(loc = 'upper right')
        self.ax.set_title(title)


    def viewSinglePCAplot(self, current_dataset = 0, plot_location = '111', title = ''):
        """
        Assemble a PCA plot
        """
        self.fig = plt.figure()

        self.assembleSubPlot(current_dataset = 0, plot_location = plot_location, title = title)
        self.ax.legend(loc = 'upper right')

        plt.show()

        # TODO:
        # ref: https://stackoverflow.com/questions/23424282/how-to-get-azimuth-and-elevation-from-a-matplotlib-figure
        self.fig.savefig(self.output_file_name, idp = 300)


    def viewSanityCheckPlots(self):
        """
        Generate plots for sanity check
        """
        if self.filter != 'None':
            self.fig = plt.figure()

            self.assembleSubPlot(current_dataset = 0, plot_location = '231', title = 'Unfiltered')
            self.assembleSubPlot(current_dataset = 1, plot_location = '232', title = 'filtered out')
            self.assembleSubPlot(current_dataset = 2, plot_location = '233', title = 'filtered')
            self.assembleSubPlot(current_dataset = 3, plot_location = '234', title = 'filtered;\nnon-infoRich')
            self.assembleSubPlot(current_dataset = 4, plot_location = '235', title = 'filtered;\ninfoRich')
            self.assembleSubPlot(current_dataset = 5, plot_location = '236', title = 'non-infoRich')

            plt.show()
            self.fig.savefig(self.output_file_name, idp = 300)

        else:
            self.fig = plt.figure()

            self.assembleSubPlot(current_dataset = 0, plot_location = '131', title = 'Unfiltered')
            self.assembleSubPlot(current_dataset = 1, plot_location = '132', title = 'non-infoRich')
            self.assembleSubPlot(current_dataset = 2, plot_location = '133', title = 'infoRich')

            plt.show()
            self.fig.savefig(self.output_file_name, idp = 300)


        # TODO: Add force_2D option
        #       Provent mixed 3D and 2D plots in sanity check plot


class RetrospectPlot:
    """
    Generate 2D or 3D PCA plot for retrospect module
    """

    def __init__(self, coordinate_and_setting_df, dimension, level, output_file_name, PC_annotation):
        self.coordinate_and_setting_df = coordinate_and_setting_df
        self.dimension = dimension
        self.level = level
        self.output_file_name = output_file_name
        self.PC_annotation = PC_annotation

    def plot(self):
        fig = plt.figure()

        if self.dimension == '3D':
            ax = fig.add_subplot(111, projection='3d')
        elif self.dimension == '2D':
            ax = fig.add_subplot(111)

        x = self.coordinate_and_setting_df['PC1']
        y = self.coordinate_and_setting_df['PC2']
        fc = list(self.coordinate_and_setting_df['fill_color'])
        ec = list(self.coordinate_and_setting_df['edge_color'])
        s = list(self.coordinate_and_setting_df['shape'])
        if self.dimension == '3D':
            z = self.coordinate_and_setting_df['PC3']

        ## why it does not work:
        #  ref: https://github.com/matplotlib/matplotlib/issues/11155
        # TODO: track this issue; for now; please manually change the setting and import by -p
        # TODO: add a WarningCode to notify user
        try:
            if list(set(s)) != ['o']:
                 raise WarningCode8(silence = False)
        except:
            pass

        if self.dimension == '3D':
            ax.scatter(x, y, z, facecolors = fc, edgecolors = ec,# marker = str(s),
                       label = self.coordinate_and_setting_df[self.level], alpha = 1)
        else:
            ax.scatter(x, y, facecolors = fc, edgecolors = ec,# marker = str(s),
                       label = self.coordinate_and_setting_df[self.level], alpha = 1)

        if isinstance(self.PC_annotation, pandas.DataFrame):
            ax.set_xlabel(self.PC_annotation['PC w percent explain'][0])
            ax.set_ylabel(self.PC_annotation['PC w percent explain'][1])
            if self.dimension == '3D':
                ax.set_zlabel(self.PC_annotation['PC w percent explain'][2])
        else:
            ax.set_xlabel('PC1')
            ax.set_ylabel('PC2')
            if self.dimension == '3D':
                ax.set_zlabel('PC3')

        plt.show()
        fig.savefig(self.output_file_name, idp = 300)



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_projection

from ..bake import BakeCommonArgs
from ..main.basic.read import RawDataImport, RetrospectDataImport
from ..posthoc.visual.projection import ProjectionArgs, projectNew
from ..toolbox.technical import flattern, emptyNumpyArray
from ..troubleshoot.err.error import *

import numpy.testing as npt
import unittest
import pandas
import numpy
import os


class TestProjectionArgs(unittest.TestCase):

    def Test_ProjectionArgs(self):
        print('\ntest_ProjectionArgs:')
        print('        case 1: error handling')
        print('             1.1: missing args.i')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode32):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.2: unexpect v input')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode34):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.3: unexpect s1 or s2')
        print('             1.3.1: unexpect s1')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode35):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.3.2: unexpect s2')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv',
                        '-s2', 'emmer/data/bake_data_dir_9']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode16):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.4: mismatch v and s1')
        print('             1.4.1: do not have exactly the same feature names')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/incorrect_colmean_1.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode42):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.4.2: feature names in -v and -s1 are not in the same order')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/incorrect_colmean_2.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode42):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.5: mismatch s1 and s2')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv',
                        '-s2', 'emmer/data/bake_data_dir_9/incorrect__data_colstd.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode48):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('             1.6: missing x')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode43):
            projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)


        print('        ---------------------------------------------------')
        print('        case 2: expect to work')
        print('             2.1: single file')
        print('             2.1.1: s1')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv',
                        '-x', 'emmer/data/bake_data_dir_9/new_observation.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                         suppress = True)

        my_result = len(projection_args.input)
        expected_result = 1
        #['emmer/data/bake_data_dir_9/new_observation.csv']

        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.1.2: s1 and s2')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv',
                        '-s2', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colstd.csv',
                        '-x', 'emmer/data/bake_data_dir_9/new_observation.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                         suppress = True)

        my_result = len(projection_args.input)
        expected_result = 1
        #['emmer/data/bake_data_dir_9/new_observation.csv']

        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.2: mulitple file')
        print('             2.2.1: get correct file names')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv',
                        '-x', 'emmer/data/bake_data_dir_9/multiple_input_file']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        projection_args = ProjectionArgs(args = processed_args.args, current_wd = current_wd,
                                         suppress = True)

        my_result = len(projection_args.input)
        expected_result = 2
        #['emmer/data/bake_data_dir_9/multiple_input_file/new_obs_case1.csv', 'emmer/data/bake_data_dir_9/multiple_input_file/new_obs_case2.csv']

        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             2.2.2: files merged correctly')
        my_result = len(flattern(projection_args.new_obs_merged[['ASV_100']].values.tolist()))
        expected_result = 30
        self.assertEqual(my_result, expected_result)

        my_result = len(flattern(projection_args.new_obs_merged[['ASV_100']].dropna().values.tolist()))
        expected_result = 13
        self.assertEqual(my_result, expected_result)

        my_result = flattern(projection_args.new_obs_merged[['ASV_100']].values.tolist())[0:12]
        expected_result = [0.004942689, 0.0, 0.0, 0.0017586879999999998, 0.0, 0.0, 0.0, 0.001431038, 0.004954967, 0.003545886, 0.002470356, 0.004484954]
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.2.3: reorder input observation before scaling')
        my_result = flattern(projection_args.new_obs_merged_ordered[['ASV_1']].values.tolist())
        expected_result = [0.194372006, 0.0, 0.003101108, 0.0, 0.046333545999999996, 0.106633061, 0.185715338, 0.013310921000000002, 0.009116045, 0.035008113,
                           0.054677207, 0.002199074, 0.10117152900000001, 0.08226267599999999, 0.013001998, 0.017784772, 0.038146321000000004, 0.04967214599999999,
                           0.062960378, 0.10306026800000001, 0.040645482999999996, 0.163720676, 0.240451627, 0.23205805100000002, 0.038100764, 0.029639332, 0.06947714,
                           0.036716487, 0.0, 0.047442625999999995]
        self.assertEqual(my_result, expected_result)

        # make sure that the dimensionality is correct
        my_result = list(projection_args.new_obs_merged_ordered.shape)
        expected_result = [len(list(projection_args.new_obs_merged.index.values)), len(list(projection_args.colmean_df.feature_names))]
        self.assertEqual(my_result, expected_result)

        # make sure they are at the save order
        my_result = list(projection_args.new_obs_merged_ordered.columns.values)
        expected_result = list(projection_args.colmean_df.feature_names)
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestProjectNew(unittest.TestCase):

    def test_projectNew(self):
        print('\ntest_projectNew:')
        print('        case 1: s1')
        print('             1.1: expect to work')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/filtered_infoRich__data_colmean.csv',
                        '-x', 'emmer/data/bake_data_dir_9/new_observation.csv',
                        '-o', 'unittest', '-r']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        projection_new = projectNew(args = processed_args.args, current_wd = current_wd, retrospect_dir = current_wd,
                                output_file_tag = processed_args.output_file_tag, suppress = False)

        print('             1.2: dimensionality')
        my_result = list(projection_new.shape)
        expected_result = [47,7]
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             1.3: check projecting result')
        result_df = projection_new[projection_new.index.str.startswith('new_observation__')][['PC1', 'PC2', 'PC3']]
        my_result = numpy.array(flattern(result_df.values.tolist()))
        expected_df = projection_new[projection_new.index.str.startswith('group_B__')][['PC1', 'PC2', 'PC3']]
        expected_result = numpy.array(flattern(expected_df.values.tolist()))
        npt.assert_almost_equal(my_result, expected_result, decimal = 3)

        os.remove('unittest__retrospect_both_new_and_original_coordinates.csv')
        os.remove('unittest__retrospect_project_new.csv')


        print('===========================================================')
        ## FIXME: Unable to use emmer.bake projection module when set emmer.harvest == T

        print('        case 2: s1 and s2')
        print('             2.1: expect to work')
        sys.argv[1:] = ['-m', 'Projection', '-i', 'emmer/data/bake_data_dir_9/s1_s2/filtered_infoRich__PCA_coordinates.csv',
                        '-v', 'emmer/data/bake_data_dir_9/s1_s2/filtered_infoRich__transformation_matrix.csv',
                        '-s1', 'emmer/data/bake_data_dir_9/s1_s2/filtered_infoRich__data_colmean.csv',
                        '-s2', 'emmer/data/bake_data_dir_9/s1_s2/filtered_infoRich__data_colstd.csv',
                        '-x', 'emmer/data/bake_data_dir_9/s1_s2/new_observations/new_obs_filterd_data.csv',
                        '-o', 'unittest', '-r']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        projection_new = projectNew(args = processed_args.args, current_wd = current_wd, retrospect_dir = current_wd,
                                    output_file_tag = processed_args.output_file_tag, suppress = False)

        result_df = projection_new[projection_new.index.str.startswith('new_obs_filterd_data__')][['PC1', 'PC2', 'PC3']]  ## TODO: Work here!!!!
        my_result = numpy.array(flattern(result_df.values.tolist()))
        expected_df = projection_new[projection_new.index.str.startswith('group_A__')][['PC1', 'PC2', 'PC3']]
        expected_result = numpy.array(flattern(expected_df.values.tolist()))
        #npt.assert_almost_equal(my_result, expected_result, decimal = 3)       ## FIXME

        os.remove('unittest__retrospect_both_new_and_original_coordinates.csv')
        os.remove('unittest__retrospect_project_new.csv')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3


from ...main.basic.read import RetrospectDataImport
from ...troubleshoot.err.error import *
from ...troubleshoot.warn.warning import *
from ...troubleshoot.inquire.input import InputCode2
from ...toolbox.recorder import UpdateNoteBook

from skbio.stats.distance import permanova
from scipy.spatial import distance_matrix
from skbio import DistanceMatrix
import pandas
import numpy
import sys
import os


"""
Support emmer.bake Permanova mode

conduct PERMANOVA test on different groups in the PCA plot
"""

class PermanovaArgs:
    """
    Take common arguments for Permanova modes when running bake modules

    Objective: so we can test and use @

    Argument:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        current_wd -- Type: str
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        silence -- Type: boolean

    Attributes:
        dist_matrix -- Type: skbio.stats.distance._base.DistanceMatrix
        group_set -- Type: set
        individual -- Type: list
        cluster -- Type: list
        coordinate_w_info -- Type: RetrospectDataImport.coordinate_w_info class object (pandas Dataframe)
                             Updated PCA coordinates with grouping information
        warning_code -- Type: str
                        for unittest

    """
    def __init__(self, args, current_wd, suppress, silence):
        # expect only args.p or args.i, not both
        if args.p:
            try:
                if args.i:
                    raise WarningCode13(silence = silence)

                parameter_df = pandas.read_csv(os.path.join(current_wd, args.p), index_col = 0, header = 0)  # TODO: FileNotFoundError  ## TODO use 'RetrospectDataImport'
                self.cluster = list(parameter_df['cluster'])

            except WarningCode13:
                self.warning_code = '13'

        else:
            try:
                if args.i:
                    print(f'Input: {args.i}\n')

                    pca_coordinate = RetrospectDataImport(file_name = os.path.join(current_wd, args.i), type = 'coordinate') # TODO: FileNotFoundError
                    self.individual = list(pca_coordinate.coordinate_w_info['individual'])
                    self.dist_matrix = DistanceMatrix(distance_matrix(pca_coordinate.coordinate_select_np, pca_coordinate.coordinate_select_np))
                    self.coordinate_w_info = pca_coordinate.coordinate_w_info
                    self.group_set = pca_coordinate.group_set
                    print(self.group_set)

                    # set parameter by interacting with users
                    cluster_map = InputCode2(set = self.group_set, suppress = suppress)
                    self.coordinate_w_info['cluster'] = self.coordinate_w_info['group'].map(cluster_map.cluster_dict)
                    self.cluster = self.coordinate_w_info['cluster']

                else:
                     raise Error(code = '8')

            except Error as e:
                raise ErrorCode8(suppress = suppress) from e


def permanovaResult(args, current_wd, retrospect_dir, output_file_tag, notebook_name, suppress, silence, neglect):
    # python3 -m emmer.bake -m 'Permanova' -i emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv

    permanova_args = PermanovaArgs(args = args, current_wd = current_wd, suppress = suppress, silence = silence)

    ## conduct PERMANOVA
    numpy.random.seed(0)

    result = permanova(permanova_args.dist_matrix, permanova_args.cluster, permutations = 999)  ## TODO: allow user-define $permutations and $seed
    print(result)

    notebook = UpdateNoteBook(notebook_name = notebook_name, neglect =  neglect).updatePermanovaResult(set_seed = '0', set_cluster = permanova_args.cluster, test_result = result)

    parameter_df = pandas.DataFrame({'individual': permanova_args.individual, 'cluster': permanova_args.cluster})
    output_file_name = os.path.join(retrospect_dir, (output_file_tag + '_retrospect_permanova_parameter.csv'))
    parameter_df.to_csv(output_file_name)
